/*
    bttv - Bt848 frame grabber driver

    Copyright (C) 1996,97,98 Ralph  Metzler (rjkm@thp.uni-koeln.de)
                           & Marcus Metzler (mocm@thp.uni-koeln.de)
    (c) 1999,2000 Gerd Knorr <kraxel@goldbach.in-berlin.de>
    V4L2 conversion
    (c) 2000 Justin Schoeman <justin@suntiger.ee.up.ac.za>

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#include <linux/version.h>
#include <linux/module.h>
#include <linux/delay.h>
#include <linux/errno.h>
#include <linux/fs.h>
#include <linux/kernel.h>
#include <linux/major.h>
#include <linux/malloc.h>
#include <linux/mm.h>
#include <linux/poll.h>
#include <linux/pci.h>
#include <linux/signal.h>
#include <asm/io.h>
#include <linux/ioport.h>
#include <asm/pgtable.h>
#include <asm/page.h>
#include <linux/sched.h>
#include <asm/segment.h>
#include <linux/types.h>
#include <linux/wrapper.h>
#include <linux/interrupt.h>
#include <linux/kmod.h>
#include <linux/vmalloc.h>
#include <linux/init.h>

#include "bttv.h"
#include "bttvp.h"
#include "tuner.h"

#define DEBUG(x)	/* Debug driver */
#define MIN(a,b) (((a)>(b))?(b):(a))
#define MAX(a,b) (((a)>(b))?(a):(b))
#define Dprintk(x,args...) if((x)<=bttv_debug) printk(##args)
#define Vprintk(x,args...) if((x)<=bttv_verbose) printk(##args)

int bttv_num;			/* number of Bt848s in use */
struct bttv bttvs[BTTV_MAX];

/* insmod args */
MODULE_PARM(triton1,"i");
MODULE_PARM_DESC(triton1,"The TV card requires triton1 bugfix, default is 0 (no)");
MODULE_PARM(radio,"1-4i");
MODULE_PARM_DESC(radio,"The TV card supports radio, default is 0 (no)");
MODULE_PARM(bigendian,"i");
MODULE_PARM_DESC(bigendian,"byte order of the framebuffer, default is native endian");
MODULE_PARM(bttv_verbose,"i");
MODULE_PARM_DESC(bttv_verbose,"verbose startup messages, default is 1 (yes)");
MODULE_PARM(bttv_gpio,"i");
MODULE_PARM_DESC(bttv_gpio,"log gpio changes, default is 0 (no)");
MODULE_PARM(bttv_debug,"i");
MODULE_PARM_DESC(bttv_debug,"debug messages, default is 0 (no)");
MODULE_PARM(irq_debug,"i");
MODULE_PARM_DESC(irq_debug,"irq handler debug messages, default is 0 (no)");

/* v4l2 specific insmod args */
MODULE_PARM(bttv_video,"1-4i");
MODULE_PARM_DESC(bttv_video,"minor number for video capture device, default is 0,1,2,3");
MODULE_PARM(bttv_vbi,"1-4i");
MODULE_PARM_DESC(bttv_vbi,"minor number for vbi capture device, default is 224,225,226,227");
MODULE_PARM(bttv_norm,"1-4i");
MODULE_PARM_DESC(bttv_norm,"default norm (video standard for each channel, default is 0,0,0,0 (PAL)");
MODULE_PARM(bttv_maxmem,"i");
MODULE_PARM_DESC(bttv_maxmem,"maximum memory to allow users to allocat for capture buffers (Meg), default is 8MB");

MODULE_DESCRIPTION("bttv - v4l2 driver module for bt848/878 based cards");
MODULE_AUTHOR("Ralph  Metzler & Marcus Metzler & Gerd Knorr - V4L2 additions by Justin Schoeman");

#if defined(__sparc__) || defined(__powerpc__)
static unsigned int bigendian=1;
#else
static unsigned int bigendian=0;
#endif
static int triton1=0;
static unsigned int radio[BTTV_MAX] = {0,0,0,0};
static unsigned int bttv_debug = 0;
static unsigned int bttv_verbose = 1;
static unsigned int irq_debug = 0;
static unsigned int bttv_gpio = 0;
static unsigned int bttv_maxmem = 8;
static int bttv_video[BTTV_MAX] = {0,1,2,3};
static int bttv_vbi[BTTV_MAX] = {224,225,226,227};
static int bttv_norm[BTTV_MAX] = {0,0,0,0};

#define I2C_TIMING (0x7<<4)
#define I2C_DELAY   10

#define I2C_SET(CTRL,DATA) \
    { btwrite((CTRL<<1)|(DATA), BT848_I2C); udelay(I2C_DELAY); }
#define I2C_GET()   (btread(BT848_I2C)&1)

#define BURSTOFFSET 76
#define BTTV_ERRORS 5

#define BTTV_MM 

#ifdef BTTV_MM

/*******************************/
/* Memory management functions */
/*******************************/

#define MDEBUG(x)	do { } while(0)		/* Debug memory management */

/* [DaveM] I've recoded most of this so that:
 * 1) It's easier to tell what is happening
 * 2) It's more portable, especially for translating things
 *    out of vmalloc mapped areas in the kernel.
 * 3) Less unnecessary translations happen.
 *
 * The code used to assume that the kernel vmalloc mappings
 * existed in the page tables of every process, this is simply
 * not guarenteed.  We now use pgd_offset_k which is the
 * defined way to get at the kernel page tables.
 */

/* Given PGD from the address space's page table, return the kernel
 * virtual mapping of the physical memory mapped at ADR.
 */
static inline unsigned long uvirt_to_kva(pgd_t *pgd, unsigned long adr)
{
        unsigned long ret = 0UL;
	pmd_t *pmd;
	pte_t *ptep, pte;
  
	if (!pgd_none(*pgd)) {
                pmd = pmd_offset(pgd, adr);
                if (!pmd_none(*pmd)) {
                        ptep = pte_offset(pmd, adr);
                        pte = *ptep;
                        if(pte_present(pte)) {
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,3,23)
                                ret  = pte_page(pte);
#else
				ret  = (unsigned long) page_address(pte_page(pte));
#endif
				ret |= (adr & (PAGE_SIZE - 1));
				
			}
                }
        }
        MDEBUG(printk("uv2kva(%lx-->%lx)", adr, ret));
	return ret;
}

static inline unsigned long uvirt_to_bus(unsigned long adr) 
{
        unsigned long kva, ret;

        kva = uvirt_to_kva(pgd_offset(current->mm, adr), adr);
	ret = virt_to_bus((void *)kva);
        MDEBUG(printk("uv2b(%lx-->%lx)", adr, ret));
        return ret;
}

static inline unsigned long kvirt_to_bus(unsigned long adr) 
{
        unsigned long va, kva, ret;

        va = VMALLOC_VMADDR(adr);
        kva = uvirt_to_kva(pgd_offset_k(va), va);
	ret = virt_to_bus((void *)kva);
        MDEBUG(printk("kv2b(%lx-->%lx)", adr, ret));
        return ret;
}

/* Here we want the physical address of the memory.
 * This is used when initializing the contents of the
 * area and marking the pages as reserved.
 */
static inline unsigned long kvirt_to_pa(unsigned long adr) 
{
        unsigned long va, kva, ret;

        va = VMALLOC_VMADDR(adr);
        kva = uvirt_to_kva(pgd_offset_k(va), va);
	ret = __pa(kva);
        MDEBUG(printk("kv2pa(%lx-->%lx)", adr, ret));
        return ret;
}

static void * rvmalloc(signed long size)
{
	void * mem;
	unsigned long adr, page;

	mem=vmalloc_32(size);
	if (mem) 
	{
		memset(mem, 0, size); /* Clear the ram out, no junk to the user */
	        adr=(unsigned long) mem;
		while (size > 0) 
                {
	                page = kvirt_to_pa(adr);
			mem_map_reserve(virt_to_page(__va(page)));
			adr+=PAGE_SIZE;
			size-=PAGE_SIZE;
		}
	}
	return mem;
}

static void rvfree(void * mem, signed long size)
{
        unsigned long adr, page;
        
	if (mem) 
	{
	        adr=(unsigned long) mem;
		while (size > 0) 
                {
	                page = kvirt_to_pa(adr);
			mem_map_unreserve(virt_to_page(__va(page)));
			adr+=PAGE_SIZE;
			size-=PAGE_SIZE;
		}
		vfree(mem);
	}
}
#else
static void * rvmalloc(signed long size)
{
	void * mem;

	mem=vmalloc_32(size);
	if (mem) 
		memset(mem, 0, size); /* Clear the ram out, no junk to the user */
	return mem;
}

static void rvfree(void * mem, signed long size)
{
	if (mem) 
		vfree(mem);
}
#endif /* BTTV_MM */

/* Some formats */

struct format_t
{
	char description[32];
	u32 pixelformat;
	u32 flags;
	u32 depth;
	int bt848fmt;
	int bpp;
	int bswap;
	int packed;
	int uvswap;
	int size_shift;
	int shift;
	int lmask;
};

struct format_t formats[] =
{
	{{"Greyscale-8"}, V4L2_PIX_FMT_GREY, V4L2_FMT_CS_601YUV, 8, BT848_COLOR_FMT_E_Y8, 8, 0, 1, 0, 0, 0, 0},
	{{"RGB-8 (3-3-2)"}, V4L2_PIX_FMT_RGB332, 0, 8, BT848_COLOR_FMT_E_RGB8, 8, 0, 1, 0, 0, 0, 0},
	{{"RGB-15 (5-5-5)"}, V4L2_PIX_FMT_RGB555, 0, 16, BT848_COLOR_FMT_E_RGB15, 15, 0, 1, 0, 0, 0, 0},
	{{"RGB-16 (5-6-5)"}, V4L2_PIX_FMT_RGB565, 0, 16, BT848_COLOR_FMT_E_RGB16, 16, 0, 1, 0, 0, 0, 0},
	{{"BGR-24 (8-8-8)"}, V4L2_PIX_FMT_BGR24, 0, 24, BT848_COLOR_FMT_E_RGB24, 24, 0, 1, 0, 0, 0, 0},
	{{"BGR-32 (8-8-8-8)"}, V4L2_PIX_FMT_BGR32, 0, 32, BT848_COLOR_FMT_E_RGB32, 32, 0, 1, 0, 0, 0, 0},
	{{"YVU410 (YVU 4:1:0)"}, V4L2_PIX_FMT_YVU410, 0, 9, BT848_COLOR_FMT_E_YCrCb411, 9, 0, 0, 0, 4, 2, 3},
	{{"YUV410 (YUV 4:1:0)"}, V4L2_PIX_FMT_YUV410, 0, 9, BT848_COLOR_FMT_E_YCrCb411, 9, 0, 0, 1, 4, 2, 3},
	{{"YVU411P (YVU 4:1:1)"}, V4L2_PIX_FMT_YVU411P, 0, 12, BT848_COLOR_FMT_E_YCrCb411, 12, 0, 0, 0, 2, 2, 0},
	{{"YVU420 (YVU 4:2:0)"}, V4L2_PIX_FMT_YVU420, 0, 12, BT848_COLOR_FMT_E_YCrCb422, 12, 0, 0, 0, 2, 1, 1},
	{{"YUV420 (YUV 4:2:0)"}, V4L2_PIX_FMT_YUV420, 0, 12, BT848_COLOR_FMT_E_YCrCb422, 12, 0, 0, 1, 2, 1, 1},
	{{"YVU422P (YVU 4:2:2)"}, V4L2_PIX_FMT_YVU422P, 0, 16, BT848_COLOR_FMT_E_YCrCb422, 16, 0, 0, 0, 1, 1, 0},
	{{"YUYV (YUV 4:2:2)"}, V4L2_PIX_FMT_YUYV, 0, 16, BT848_COLOR_FMT_E_YUY2, 16, 0, 1, 0, 0, 0, 0},
	{{"UYVY (YVU 4:2:2)"}, V4L2_PIX_FMT_UYVY, 0, 16, BT848_COLOR_FMT_E_YUY2, 16, 1, 1, 0, 0, 0, 0}
};

int format_max=14;

int find_format(u32 pfmt)
{
	int i;
	for(i=0; i<format_max; i++)
		if(formats[i].pixelformat==pfmt)return i;
	return -1;
}

/* init + register i2c algo-bit adapter */
static int __devinit init_bttv_i2c(struct bttv *btv)
{
	memcpy(&btv->i2c_adap, &bttv_i2c_adap_template, sizeof(struct i2c_adapter));
	memcpy(&btv->i2c_algo, &bttv_i2c_algo_template, sizeof(struct i2c_algo_bit_data));
	memcpy(&btv->i2c_client, &bttv_i2c_client_template, sizeof(struct i2c_client));

/*	sprintf(btv->i2c_adap.name+strlen(btv->i2c_adap.name),
		" #%d", btv->nr); */
	sprintf(btv->i2c_adap.name, "bttv%d: i2c adapter", btv->nr);
        btv->i2c_algo.data = btv;
        btv->i2c_adap.data = btv;
        btv->i2c_adap.algo_data = &btv->i2c_algo;
        btv->i2c_client.adapter = &btv->i2c_adap;

	bttv_bit_setscl(btv,1);
	bttv_bit_setsda(btv,1);

	btv->i2c_rc = i2c_bit_add_bus(&btv->i2c_adap);
	return btv->i2c_rc;
}


/* ----------------------------------------------------------------------- */

void bttv_gpio_tracking(struct bttv *btv, char *comment)
{
	unsigned int outbits, data;
	outbits = btread(BT848_GPIO_OUT_EN);
	data    = btread(BT848_GPIO_DATA);
	Dprintk(2,KERN_DEBUG "bttv%d: gpio: en=%08x, out=%08x in=%08x [%s]\n",
	       btv->nr,outbits,data & outbits, data & ~outbits, comment);
}

static char *audio_modes[] = { "audio: tuner", "audio: radio", "audio: extern",
			       "audio: intern", "audio: off" };

static void audio(struct bttv *btv, int mode, int no_irq_context)
{
	btaor(bttv_tvcards[btv->type].gpiomask, ~bttv_tvcards[btv->type].gpiomask,
              BT848_GPIO_OUT_EN);

	switch (mode)
	{
	        case AUDIO_MUTE:
                        btv->audio|=AUDIO_MUTE;
			break;
 		case AUDIO_UNMUTE:
			btv->audio&=~AUDIO_MUTE;
			mode=btv->audio;
			break;
		case AUDIO_OFF:
			mode=AUDIO_OFF;
			break;
		case AUDIO_ON:
			mode=btv->audio;
			break;
		default:
			btv->audio&=AUDIO_MUTE;
			btv->audio|=mode;
			break;
	}
        /* if audio mute or not in H-lock, turn audio off */
	if ((btv->audio&AUDIO_MUTE))
	        mode=AUDIO_OFF;
        if ((mode == AUDIO_TUNER) && (btv->radio))
		mode = AUDIO_RADIO;
	btaor(bttv_tvcards[btv->type].audiomux[mode],
		~bttv_tvcards[btv->type].gpiomask, BT848_GPIO_DATA);
	if (bttv_gpio)
		bttv_gpio_tracking(btv,audio_modes[mode]);
	if (no_irq_context)
		bttv_call_i2c_clients(btv,AUDC_SET_INPUT,&(mode));
}




/* If Bt848a or Bt849, use PLL for PAL/SECAM and crystal for NTSC*/

/* Frequency = (F_input / PLL_X) * PLL_I.PLL_F/PLL_C 
   PLL_X = Reference pre-divider (0=1, 1=2) 
   PLL_C = Post divider (0=6, 1=4)
   PLL_I = Integer input 
   PLL_F = Fractional input 
   
   F_input = 28.636363 MHz: 
   PAL (CLKx2 = 35.46895 MHz): PLL_X = 1, PLL_I = 0x0E, PLL_F = 0xDCF9, PLL_C = 0
*/

static void set_pll_freq(struct bttv *btv, unsigned int fin, unsigned int fout)
{
        unsigned char fl, fh, fi;
        
        /* prevent overflows */
        fin/=4;
        fout/=4;

        fout*=12;
        fi=fout/fin;

        fout=(fout%fin)*256;
        fh=fout/fin;

        fout=(fout%fin)*256;
        fl=fout/fin;

        /*printk("0x%02x 0x%02x 0x%02x\n", fi, fh, fl);*/
        btwrite(fl, BT848_PLL_F_LO);
        btwrite(fh, BT848_PLL_F_HI);
        btwrite(fi|BT848_PLL_X, BT848_PLL_XCI);
}

static int set_pll(struct bttv *btv)
{
        int i;
	unsigned long tv;

        if (!btv->pll.pll_crystal)
                return 0;

        if (btv->pll.pll_ifreq == btv->pll.pll_ofreq) {
                /* no PLL needed */
                if (btv->pll.pll_current == 0) {
                        /* printk ("bttv%d: PLL: is off\n",btv->nr); */
                        return 0;
                }
                Dprintk (2,"bttv%d: PLL: switching off\n",btv->nr);
                btwrite(0x00,BT848_TGCTRL);
                btwrite(0x00,BT848_PLL_XCI);
                btv->pll.pll_current = 0;
                return 0;
        }

        if (btv->pll.pll_ofreq == btv->pll.pll_current) {
                /* printk("bttv%d: PLL: no change required\n",btv->nr); */
                return 1;
        }

	Vprintk(1, "bttv%d: PLL: %d => %d ... ",btv->nr,
		       btv->pll.pll_ifreq, btv->pll.pll_ofreq);

	set_pll_freq(btv, btv->pll.pll_ifreq, btv->pll.pll_ofreq);

	/*  Let other people run while the PLL stabilizes */
	tv=jiffies+HZ/10;       /* .1 seconds */
	do
	{
		schedule();
	}
	while(time_before(jiffies,tv));

        for (i=0; i<100; i++) 
        {
                if ((btread(BT848_DSTATUS)&BT848_DSTATUS_PLOCK))
                        btwrite(0,BT848_DSTATUS);
                else
                {
                        btwrite(0x08,BT848_TGCTRL);
                        btv->pll.pll_current = btv->pll.pll_ofreq;
			Vprintk(1,"ok\n");
                        return 1;
                }
                mdelay(10);
        }
        btv->pll.pll_current = 0;
	Vprintk(1,"oops\n");
        return -1;
}

static void bt848_muxsel(struct bttv *btv, unsigned int input)
{
#if 0 /* seems no card uses this ... */
	btaor(bttv_tvcards[btv->type].gpiomask2,~bttv_tvcards[btv->type].gpiomask2,
              BT848_GPIO_OUT_EN);
#endif

	/* This seems to get rid of some synchronization problems */
	btand(~(3<<5), BT848_IFORM);
	mdelay(10); 
        
	input %= bttv_tvcards[btv->type].video_inputs;
	if (input==bttv_tvcards[btv->type].svhs) 
	{
		btor(BT848_CONTROL_COMP, BT848_E_CONTROL);
		btor(BT848_CONTROL_COMP, BT848_O_CONTROL);
	}
	else
	{
		btand(~BT848_CONTROL_COMP, BT848_E_CONTROL);
		btand(~BT848_CONTROL_COMP, BT848_O_CONTROL);
	}
	btaor((bttv_tvcards[btv->type].muxsel[input&7]&3)<<5, ~(3<<5), BT848_IFORM);
	audio(btv, (input!=bttv_tvcards[btv->type].tuner) ? 
              AUDIO_EXTERN : AUDIO_TUNER, 1);
#if 0 /* seems no card uses this ... */
	btaor(bttv_tvcards[btv->type].muxsel[input]>>4,
		~bttv_tvcards[btv->type].gpiomask2, BT848_GPIO_DATA);
	if (bttv_gpio)
		bttv_gpio_tracking(btv,"muxsel");
#endif
}

static void bt848_set_input(struct bttv *btv, unsigned int input)
{
	struct video_channel vc;
	bttv_call_i2c_clients(btv, VIDIOCGCHAN, &vc);
	vc.norm =  btv->norm;
	bttv_call_i2c_clients(btv, VIDIOCSCHAN, &vc);
	bt848_muxsel(btv, input);
	btv->input = input;
}

struct tvnorm 
{
	int v4l2_id;
        u32 Fsc;
        u16 swidth, maxwidth, sheight; /* scaled standard width, height */
	u16 totalwidth;
	u8 adelay, bdelay, iform;
	u32 scaledtwidth;
	u16 hdelayx1, hactivex1;
	u16 vdelay;
        u8 vbipack;
};

static struct tvnorm tvnorms[] = {
	/* PAL-BDGHI */
        /* max. active video is actually 922, but 924 is divisible by 4 and 3! */
 	/* actually, max active PAL with HSCALE=0 is 948, NTSC is 768 - nil */
        { V4L2_STD_PAL, 35468950,
          924, 768, 576, 1135, 0x7f, 0x72, (BT848_IFORM_PAL_BDGHI|BT848_IFORM_XT1),
          1135, 186, 924, 0x20, 255},

	/* NTSC */
	{ V4L2_STD_NTSC, 28636363,
          768, 640, 480,  910, 0x68, 0x5d, (BT848_IFORM_NTSC|BT848_IFORM_XT0),
          910, 128, 910, 0x1a, 144},
#if 0
	/* SECAM EAST */
	{ V4L2_STD_SECAM, 35468950, 
          768, 768, 576, 1135, 0x7f, 0xb0, (BT848_IFORM_SECAM|BT848_IFORM_XT1),
	  944, 186, 922, 0x20, 255},
#else
	/* SECAM L */
        { V4L2_STD_SECAM, 35468950,
          924, 768, 576, 1135, 0x7f, 0xb0, (BT848_IFORM_SECAM|BT848_IFORM_XT1),
          1135, 186, 922, 0x20, 255},
#endif
        /* PAL-NC */
        { V4L2_STD_PAL_60, 28636363,
          640, 640, 576,  910, 0x68, 0x5d, (BT848_IFORM_PAL_NC|BT848_IFORM_XT0),
          780, 130, 734, 0x1a, 144},
	/* PAL-M */
	{ V4L2_STD_PAL_M, 28636363, 
          640, 640, 480, 910, 0x68, 0x5d, (BT848_IFORM_PAL_M|BT848_IFORM_XT0),
	  780, 135, 754, 0x1a, 144},
	/* PAL-N */
	{ V4L2_STD_PAL_N, 35468950, 
          768, 768, 576, 1135, 0x7f, 0x72, (BT848_IFORM_PAL_N|BT848_IFORM_XT1),
	  944, 186, 922, 0x20, 144},
	/* NTSC-Japan */
	{ V4L2_STD_NTSC_N, 28636363,
          640, 640, 480,  910, 0x68, 0x5d, (BT848_IFORM_NTSC_J|BT848_IFORM_XT0),
	  780, 135, 754, 0x16, 144},
};
#define TVNORMS (sizeof(tvnorms)/sizeof(struct tvnorm))

/******************************/
/* bt848 functions            */
/******************************/
 
 /* Build a scatter list */
static int bm_build_scatter_list(struct bttvc *btvc, unsigned char *buffer, struct scatter_node **plist)
{
	struct bttvo *btvo = btvc->open;
	struct bttv *btv = btvo->dev;
	struct scatter_node	*list;
	int			i, n;
	unsigned char		*a;

	if (buffer == NULL)
		return -1;
	list = *plist;
	if (list == NULL)
	{
		printk("bttv%d: error - user app must kmalloc() dma table\n", btv->nr);
		return -1;
	}

	/*  Simple algorithm will just map the buffer contiguously by pages */
	/*  Note: capture_buffer is vmalloc()ed, so it's page-aligned */

	n = btvo->cap_bsize / PAGE_SIZE;
	a = buffer;
	for (i = 0; i < n; ++i)
	{
#ifdef BTTV_MM
		list[i].addr = kvirt_to_bus((unsigned long)a);
#else
		list[i].addr = v4l2_vmalloc_to_bus(a);
#endif /* BTTV_MM */
		list[i].len = PAGE_SIZE;
		a += PAGE_SIZE;
	}
	/* Last page might not be full */
	if (btvo->cap_bsize < n * PAGE_SIZE)
		list[n - 1].len = btvo->cap_bsize - (n - 1) * PAGE_SIZE;

	list[n - 1].len |= END_OF_SCATTER_LIST;

	*plist = list;
	return 0;/* ok */
}


/* Geometry calculations */

/*
 *	Set the registers for the size we have specified. Don't bother
 *	trying to understand this without the BT848 manual in front of
 *	you [AC]. 
 *
 *	PS: The manual is free for download in .pdf format from 
 *	www.brooktree.com - nicely done those folks.
 */

static void bt848_calc_geo(struct bttv *btv, struct geo_registers *geo, int width, int height, int pfmt)
{
	struct tvnorm *tvn;
	int inter;
	u32 xsf, sr;
	u16 hscale, vscale;
	int i;

	tvn=&tvnorms[btv->norm];
	inter = (height>tvn->sheight/2) ? 0:1;

	xsf = (width*tvn->scaledtwidth)/tvn->swidth;
	hscale = ((tvn->totalwidth*4096UL)/xsf-4096);
	geo->hdelay =  tvn->hdelayx1;
	geo->hdelay =  (geo->hdelay*width)/tvn->swidth;
	geo->hdelay &= 0x3fe;
	sr=((tvn->sheight>>inter)*512)/height-512;
	vscale=(0x10000UL-sr)&0x1fff;
	geo->crop=((width>>8)&0x03)|((geo->hdelay>>6)&0x0c)|
		((tvn->sheight>>4)&0x30)|((tvn->vdelay>>2)&0xc0);
	vscale |= inter ? (BT848_VSCALE_INT<<8) : 0;

#if 0
	/* Some people say interpolation looks bad ... */
	vtc = (width < 193) ? 2 : ((width < 385) ? 1 : 0);
	if (width < 767)
		geo->comb=1;
	else
		geo->comb=0;
#else
	geo->vtc = 0;
	geo->comb = 0;
#endif
	geo->hscale_lo = hscale&0xff;
	geo->hscale_hi = (hscale>>8)&0xff;
	geo->vscale_lo = vscale&0xff;
	geo->vscale_hi = (vscale>>8)&0xff;
	geo->vactive = tvn->sheight&0xff;
	geo->hactive = width&0xff;
	geo->vdelay = tvn->vdelay&0xff;

	i = find_format(pfmt);
	if(i<0)
	{
		Dprintk(1,"bttv%d: Invalid pixelformat defaulting to grey.\n", btv->nr);
		i = 0;
	}
	geo->format = formats[i].bt848fmt;
	geo->bswap = formats[i].bswap;
	if(formats[i].packed==0)
	{
		geo->comb=1;
	}
	if(bigendian && geo->format == BT848_COLOR_FMT_E_RGB32)
	{
		geo->bswap^=1; // bswap
		geo->bswap^=4; //wswap
	} else if(bigendian &&
		(geo->format == BT848_COLOR_FMT_E_RGB16 || 
		 geo->format == BT848_COLOR_FMT_E_RGB15))
	{
		geo->bswap^=1; //bswap
	}
	geo->uvswap = formats[i].uvswap;
}	

static void bt848_set_geo(struct bttv *btv, struct geo_registers *geo, int odd)
{
	int off = odd ? 0x80 : 0x00;
	
	btwrite(geo->vtc, BT848_E_VTC+off);
	btwrite(geo->hscale_hi, BT848_E_HSCALE_HI+off);
	btwrite(geo->hscale_lo, BT848_E_HSCALE_LO+off);
	btaor(geo->vscale_hi, 0xe0, BT848_E_VSCALE_HI+off);
	btwrite(geo->vscale_lo, BT848_E_VSCALE_LO+off);
	btwrite(geo->hactive, BT848_E_HACTIVE_LO+off);
	btwrite(geo->hdelay&0xff, BT848_E_HDELAY_LO+off);
	btwrite(geo->vactive, BT848_E_VACTIVE_LO+off);
	btwrite(geo->vdelay, BT848_E_VDELAY_LO+off);
	btwrite(geo->crop, BT848_E_CROP+off);
	btaor(geo->uvswap, 0xef, BT848_E_CONTROL+off);
	if(geo->comb)
		btor(BT848_VSCALE_COMB, BT848_E_VSCALE_HI+off);
	else
		btand(~BT848_VSCALE_COMB, BT848_E_VSCALE_HI+off);
	if(odd)
	{
		btaor(geo->format<<4, 0x0f, BT848_COLOR_FMT);
		btaor(geo->bswap<<1, 0xf5, BT848_COLOR_CTL);
	}
	else
	{
		btaor(geo->format, 0xf0, BT848_COLOR_FMT);
		btaor(geo->bswap, 0xfa, BT848_COLOR_CTL);
	}
}

static int capture_set_format(struct bttvo *btvo, struct v4l2_pix_format *f);

void bt848_set_standard(struct bttv *btv)
{
	struct tvnorm *tvn;
	struct v4l2_pix_format f;
	struct v4l2_standard s;
	int i;
	
	Dprintk(2,"bttv%d: Setting standard to %d\n", btv->nr, btv->norm);
	if(btv->dma)
	{
		printk("bttv%d: Cannot set standard while DMA running - aborting.\n", btv->nr);
		return;
	}
	if(btv->norm>=TVNORMS)
	{
		printk("bttv%d: invalid standard, defaulting to PAL.\n", btv->nr);
		btv->norm = 0;
	}
	tvn=&tvnorms[btv->norm];
	btwrite(tvn->adelay, BT848_ADELAY);
	btwrite(tvn->bdelay, BT848_BDELAY);
	btv->vbi_bpl = (256 + tvn->vbipack) * 4;
	btwrite(tvn->vbipack, BT848_VBI_PACK_SIZE);
	btwrite(1, BT848_VBI_PACK_DEL);
	btaor(tvn->iform,~(BT848_IFORM_NORM|BT848_IFORM_XTBOTH), BT848_IFORM);
        btv->pll.pll_ofreq = tvn->Fsc;
	set_pll(btv);
	/* Set up the dummy */
	btv->dummy_width = tvn->swidth/4;
	btv->dummy_height = tvn->sheight/4;
	btv->dummy_pfmt = V4L2_PIX_FMT_GREY;
	bt848_calc_geo(btv, &btv->dummy_geo, btv->dummy_width, btv->dummy_height, btv->dummy_pfmt);
	memset(&f, 0, sizeof(struct v4l2_pix_format));
	f.width = btv->dummy_width;
	f.height = btv->dummy_height;
	f.pixelformat = btv->dummy_pfmt;
	v4l2_video_std_construct(&s, tvn->v4l2_id, 0);
	btv->tpf = v4l2_video_std_tpf(&s);
	for(i=0; i<MAX_OPENS; i++)
		if(btv->open_data[i].isopen)
			capture_set_format(&btv->open_data[i], &f);
	for(i=0; i<MAX_VBI_OPENS; i++)
		if(btv->vbi_open_data[i].isopen)
			btv->vbi_open_data[i].vbi_fmt.sampling_rate = 0;
};


#if 0 /* perhaps I should implement this sometime ... */

/*
 * called from irq handler on fatal errors.  Takes the grabber chip
 * offline, flag it needs a reinitialization (which can't be done
 * from irq context) and wake up all sleeping proccesses.  They would
 * block forever else.  We also need someone who actually does the
 * reinitialization from process context...
 */
static void bt848_offline(struct bttv *btv)
{
	int i;
	spin_lock(&btv->s_lock);

	/* cancel all outstanding grab requests */
	btv->gq_in = 0;
	btv->gq_out = 0;
	btv->gq_grab = -1;
	for (i = 0; i < gbuffers; i++)
		if (btv->gbuf[i].stat == GBUFFER_GRABBING)
			btv->gbuf[i].stat = GBUFFER_ERROR;

	/* disable screen overlay and DMA */
	btv->risc_cap_odd  = 0;
	btv->risc_cap_even = 0;
	bt848_set_risc_jmps(btv,0);

	/* flag the chip needs a restart */
	btv->needs_restart = 1;
	spin_unlock(&btv->s_lock);

	wake_up_interruptible(&btv->waitq);
}

static void bt848_restart(struct bttv *btv)
{
 	unsigned long irq_flags;

	Vprintk(1,"bttv%d: resetting chip\n",btv->nr);
	btwrite(0xfffffUL, BT848_INT_STAT);
	btand(~15, BT848_GPIO_DMA_CTL);
	btwrite(0, BT848_SRESET);
	btwrite(virt_to_bus(btv->risc_jmp+2),
		BT848_RISC_STRT_ADD);

	/* enforce pll reprogramming */
	btv->pll.pll_current = 0;
	set_pll(btv);

	spin_lock_irqsave(&btv->s_lock, irq_flags);
	btv->errors = 0;
	btv->needs_restart = 0;
	bt848_set_geo(btv,0);
//	bt848_set_risc_jmps(btv,-1);
	spin_unlock_irqrestore(&btv->s_lock, irq_flags);
}

#endif

/* DMA FUNCTIONS HERE */
#define RISCMEM_LEFT(a, b, c) (((u32)c) - (((u32)a) - ((u32)b)))

static int maxdma(struct bttv *btv, u32 addr, struct scatter_node *list, u32 *phys)
{
	int i = 0;
	
	while (addr >= (list[i].len&~END_OF_SCATTER_LIST))
	{
		if(list[i].len&END_OF_SCATTER_LIST)
		{
			printk("bttv%d: attempt to find addr above scatter_list.\n", btv->nr);
			return -1;
		} else
		{
			addr-=list[i].len&~END_OF_SCATTER_LIST;
			i++;
		}
	}
	*phys = list[i].addr + addr;
	return ((list[i].len&~END_OF_SCATTER_LIST) - addr);
}

int bt848_krealloc(struct hw_buf *hw, u32 **re, u32 **ro)
{
	int i;
	void *newmem_e;
	void *newmem_o;
	u32 reo, roo;
	
	for(i=1; i<6; i++)
		if((PAGE_SIZE<<i)>hw->risc_len)
			break;
	if(i>5)
	{
		printk("bttv-krealloc(): request too large!\n");
		return -1;
	}
	Dprintk(2,"bttv-krealloc(): request %ld (%d)!\n", PAGE_SIZE<<i, i);
	newmem_e = kmalloc(PAGE_SIZE<<i, GFP_KERNEL);
	if(!newmem_e)
		return -1;
	newmem_o = kmalloc(PAGE_SIZE<<i, GFP_KERNEL);
	if(!newmem_o)
	{
		kfree(newmem_e);
		return -1;
	}
	reo = *re - hw->risc_even;
	roo = *ro - hw->risc_odd;
	memcpy(newmem_e, hw->risc_even, PAGE_SIZE<<(i-1));
	kfree(hw->risc_even);
	hw->risc_even = newmem_e;
	hw->risc_even_bus = cpu_to_le32(virt_to_bus(hw->risc_even));
	memcpy(newmem_o, hw->risc_odd, PAGE_SIZE<<(i-1));
	kfree(hw->risc_odd);
	hw->risc_odd = newmem_o;
	hw->risc_odd_bus = cpu_to_le32(virt_to_bus(hw->risc_odd));
	hw->risc_len = PAGE_SIZE<<i;
	*re = hw->risc_even + reo;
	*ro = hw->risc_odd + roo;
	return 0;
}

/* Make planar RISC tables */
inline int make_prisctab(struct bttv *btv, struct hw_buf *hw, int width, int height, u32 pfmt, u32 stride, int i)
{
        u32 *re, *ro, **rp, rcmd, todo;
        int inter, line, csize, bl = 0, blcr = 0, blcb = 0;
	u32 yadr = 0, cbadr = 0, cradr = 0;
	u32 yadr_phys, cbadr_phys, cradr_phys;

	inter = (height > tvnorms[btv->norm].sheight/2)?1:0;
	csize=(width*height)>>formats[i].size_shift;

	re = hw->risc_even;
	ro = hw->risc_odd;
	        
	*(ro++) = cpu_to_le32(BT848_RISC_SYNC|BT848_FIFO_STATUS_FM3);
	*(ro++) = cpu_to_le32(0);
	*(re++) = cpu_to_le32(BT848_RISC_SYNC|BT848_FIFO_STATUS_FM3);
	*(re++) = cpu_to_le32(0);
	
	Dprintk(2,"bttv%d: building planar risc buffer %dx%d %s.\n", btv->nr, width, height, formats[i].description);
	
	for(line=0; line < (height<<(1^inter)); line++)
	{
		if((line == height)||(line == 0))
		{
			yadr = 0;
			if(formats[i].uvswap)
			{
				cbadr = width*height;
				cradr = cbadr + csize;
			} else
			{
				cradr = width*height;
				cbadr = cradr + csize;
			}
		}
		if (inter)
			rp = (line&1) ? &re : &ro;
		else
			rp = (line>=height) ? &re : &ro;

		if(hw->dma->addr!=0)
		{
			if(line&formats[i].lmask)
				rcmd=BT848_RISC_WRITE1S23|BT848_RISC_SOL;
			else
				rcmd=BT848_RISC_WRITE123|BT848_RISC_SOL;
		} else
		{
			rcmd=BT848_RISC_SKIP123|BT848_RISC_SOL;
		}
		
		todo = width;
		while(todo)
		{
			if((RISCMEM_LEFT(re, hw->risc_even, hw->risc_len)<48)
			   ||(RISCMEM_LEFT(ro, hw->risc_odd, hw->risc_len)<48))
			{
				if(bt848_krealloc(hw, &re, &ro)<0)
				{
					printk("bttv%d: Uh-Oh, out of space in risc table for CAP\n", btv->nr);
					goto bugout;
				}
			}
			bl = maxdma(btv, yadr, hw->dma, &yadr_phys);
			if(bl<=0)
			{
				printk("bttv%d: Whups - scatterlist stuffed at line %d\n", btv->nr, line);
				goto bugout;
			}
			if((rcmd&(15<<28))==BT848_RISC_WRITE123)
			{
				blcr = maxdma(btv, cradr, hw->dma, &cradr_phys);
				blcb = maxdma(btv, cbadr, hw->dma, &cbadr_phys);
				if((blcr<=0)||(blcb<=0))
				{
					printk("bttv%d: Whups - scatterlist stuffed at line %d\n", btv->nr, line);
					goto bugout;
				}
				blcr<<=formats[i].shift;
				blcb<<=formats[i].shift;
				bl=(blcr<bl) ? blcr : bl;
				bl=(blcb<bl) ? blcb : bl;
			} 
			bl=(bl>todo) ? todo : bl;
			blcr=bl>>formats[i].shift;
			blcb=blcr;
			todo-=bl;
			if(!todo) rcmd|=BT848_RISC_EOL;
			
			*((*rp)++) = cpu_to_le32(rcmd|bl);
			*((*rp)++) = cpu_to_le32(blcb|(blcr<<16));
			if((rcmd&(15<<28))!=BT848_RISC_SKIP123)
				*((*rp)++) = cpu_to_le32(yadr_phys);
			yadr+=bl;
			if((rcmd&(15<<28))==BT848_RISC_WRITE123)
			{
				*((*rp)++) = cpu_to_le32(cbadr_phys);
				cbadr+=blcb;
				*((*rp)++) = cpu_to_le32(cradr_phys);
				cradr+=blcr;
			}
			rcmd&=~BT848_RISC_SOL;
		}
	}
bugout:
	Dprintk(2,"bttv%d: risc mem left: %d %d\n", btv->nr, RISCMEM_LEFT(ro, hw->risc_odd, hw->risc_len), RISCMEM_LEFT(re, hw->risc_even, hw->risc_len));
	*(ro++) = cpu_to_le32(BT848_RISC_JUMP);
	*(ro++) = cpu_to_le32(btv->cap_odd_end);
	*(re++) = cpu_to_le32(BT848_RISC_JUMP);
	*(re++) = cpu_to_le32(btv->cap_even_end);
	return 0;
}

inline int make_clip_risc(struct bttv *btv, struct hw_buf *hw, int width, int height, u32 pfmt, u32 stride, u8 *bitmap, int i)
{
        u32 *re, *ro, **rp, bpl, rcmd, addr, bl, maxaddr, minaddr;
        int inter, line, fbit;
	u32 x, cwidth, ofs, bpp;
	u8 mask = 0x80;
	
	inter = (height > tvnorms[btv->norm].sheight/2)?1:0;
	fbit = 0;
	ofs = 0;
	
	bpp = formats[i].depth;
	bpl = (width * bpp + 7) >> 3;
	minaddr = (u32)btv->fb.base[0];
	maxaddr = minaddr + (btv->fb.fmt.height * stride);
	addr = minaddr + (stride * btv->win.y);
	if(btv->win.x>0)
		addr += ((btv->win.x * bpp + 7) >> 3);
	else
		addr -= (((-btv->win.x) * bpp + 7) >> 3);

	if(!bitmap)
	{
		printk("bttv%d: No clipmap given!\n", btv->nr);
		return -1;
	}

	re = hw->risc_even;
	ro = hw->risc_odd;
        	
	*(ro++) = cpu_to_le32(BT848_RISC_SYNC|BT848_FIFO_STATUS_FM1);
	*(ro++) = cpu_to_le32(0);
	*(re++) = cpu_to_le32(BT848_RISC_SYNC|BT848_FIFO_STATUS_FM1);
	*(re++) = cpu_to_le32(0);
	
	Dprintk(2,"bttv%d: building packed risc buffer (clipped) %dx%d %s.\n", btv->nr, width, height, formats[i].description);
	
	for(line=0; line < (height<<(1^inter)); line++)
	{
		if(line == height)
		{
			addr = minaddr + (stride * btv->win.y);
			if(btv->win.x>0)
				addr += ((btv->win.x * bpp + 7) >> 3);
			else
				addr -= (((-btv->win.x) * bpp + 7) >> 3);
			ofs = 0;
			mask = 0x80;
		}
		if (inter)
			rp = (line&1) ? &re : &ro;
		else
			rp = (line>=height) ? &re : &ro;
		rcmd = BT848_RISC_SOL;
		x = 0;
		while(x<width)
		{
			if((RISCMEM_LEFT(re, hw->risc_even, hw->risc_len)<32)
			   ||(RISCMEM_LEFT(ro, hw->risc_odd, hw->risc_len)<32))
			{
				if(bt848_krealloc(hw, &re, &ro)<0)
				{
					printk("bttv%d: Uh-Oh, out of space in risc table for PV\n", btv->nr);
					goto bugout;
				}
			}
			
			fbit = bitmap[ofs]&mask;

			cwidth = 0;
			while(cwidth<(width - x))
			{
				cwidth++;
				mask>>=1;
				if(mask==0)
				{
					mask = 0x80;
					ofs++;
				}
				if((bitmap[ofs]&mask)&&(!(fbit)))break;
				if((!(bitmap[ofs]&mask))&&(fbit))break;
			}
			x+=cwidth;
			bl = (cwidth * bpp + 7) >> 3;
			
			if(x>=width) rcmd|=BT848_RISC_EOL;
			if((fbit==0)&&(hw->dma->addr!=0))
			{
				if(((addr+bl)>maxaddr)||(addr<minaddr))
				{
					printk("bttv%d: attempting to DMA beyond FB %08X %08X %08X %08X\n", btv->nr, minaddr, addr, addr+bl, maxaddr);
					goto bugout;
				}
				*((*rp)++) = cpu_to_le32(BT848_RISC_WRITE|rcmd|bl);
				*((*rp)++) = cpu_to_le32(addr);
			} else
			{
				*((*rp)++) = cpu_to_le32(BT848_RISC_SKIP|rcmd|bl);
			}
			addr+=bl;
			rcmd&=~BT848_RISC_SOL;
		}
		addr+=(stride - bpl);
	}

bugout:
	Dprintk(2,"bttv%d: risc mem left: %d %d\n", btv->nr, RISCMEM_LEFT(ro, hw->risc_odd, hw->risc_len), RISCMEM_LEFT(re, hw->risc_even, hw->risc_len));
	*(ro++) = cpu_to_le32(BT848_RISC_JUMP);
	*(ro++) = cpu_to_le32(btv->cap_odd_end);
	*(re++) = cpu_to_le32(BT848_RISC_JUMP);
	*(re++) = cpu_to_le32(btv->cap_even_end);

	return 0;
}

static int make_risctab(struct bttv *btv, struct hw_buf *hw, int width, int height, u32 pfmt, u32 stride, void *bitmap)
{
        u32 *re, *ro, **rp, rcmd, todo, addr = 0, addr_phys, bpl;
        int inter, line, bl, i;
        
        i = find_format(pfmt);
        if(i<0)
        {
        	printk("bttv%d: make risctab - invalid format - default to GREY.\n", btv->nr);
        	i=0;
        }
	inter = (height > tvnorms[btv->norm].sheight/2)?1:0;
	bpl = (width * formats[i].depth + 7) >> 3;
	if(stride==0)
		stride = bpl;
	
	if((!hw->risc_even)||(!hw->risc_odd))
	{
		hw->risc_len = PAGE_SIZE;
		hw->risc_even = (u32 *)kmalloc(hw->risc_len, GFP_KERNEL);
		hw->risc_odd = (u32 *)kmalloc(hw->risc_len, GFP_KERNEL);
	}
	if((!hw->risc_even)||(!hw->risc_odd))
	{
		printk("bttv%d: could not get memory for risc tables.\n", btv->nr);
		if(hw->risc_even)
		{
			kfree(hw->risc_even);
			hw->risc_even = NULL;
		}
		if(hw->risc_odd)
		{
			kfree(hw->risc_odd);
			hw->risc_odd = NULL;
		}
		return -1;
	}
	hw->risc_even_bus = cpu_to_le32(virt_to_bus(hw->risc_even));
	hw->risc_odd_bus = cpu_to_le32(virt_to_bus(hw->risc_odd));
	
	if((formats[i].packed)&&(bitmap))
		return make_clip_risc(btv, hw, width, height, pfmt, stride, (u8 *)bitmap, i);
	if(!formats[i].packed)
		return  make_prisctab(btv, hw, width, height, pfmt, stride, i);
	
	re = hw->risc_even;
	ro = hw->risc_odd;
        	
	*(ro++) = cpu_to_le32(BT848_RISC_SYNC|BT848_FIFO_STATUS_FM1);
	*(ro++) = cpu_to_le32(0);
	*(re++) = cpu_to_le32(BT848_RISC_SYNC|BT848_FIFO_STATUS_FM1);
	*(re++) = cpu_to_le32(0);
	
	Dprintk(2,"bttv%d: building packed risc buffer %dx%d %s.\n", btv->nr, width, height, formats[i].description);
	
	for(line=0; line < (height<<(1^inter)); line++)
	{
		if(line == height)
			addr = 0;
		if (inter)
			rp = (line&1) ? &re : &ro;
		else
			rp = (line>=height) ? &re : &ro;
		if(hw->dma->addr!=0)
			rcmd = BT848_RISC_WRITE|BT848_RISC_SOL;
		else
			rcmd = BT848_RISC_SKIP|BT848_RISC_SOL;
		todo = bpl;
		while(todo)
		{
			if((RISCMEM_LEFT(re, hw->risc_even, hw->risc_len)<32)
			   ||(RISCMEM_LEFT(ro, hw->risc_odd, hw->risc_len)<32))
			{
				if(bt848_krealloc(hw, &re, &ro)<0)
				{
					printk("bttv%d: Uh-Oh, out of space in risc table for CAP\n", btv->nr);
					goto bugout;
				}
			}
			bl = maxdma(btv, addr, hw->dma, &addr_phys);
			if(bl<=0)
			{
				printk("bttv%d: Whups - scatterlist stuffed at line %d\n", btv->nr, line);
				goto bugout;
			}
			bl = (todo>bl)?bl:todo;
			todo-=bl;
			if(!todo)
			{
				rcmd|=BT848_RISC_EOL;
				addr+=(stride-bpl);
			}
			*((*rp)++) = cpu_to_le32(rcmd|bl);
			if(hw->dma->addr!=0)
				*((*rp)++) = cpu_to_le32(addr_phys);
			addr+=bl;
			rcmd&=~BT848_RISC_SOL;
		}
	}
bugout:
	Dprintk(2,"bttv%d: risc mem left: %d %d\n", btv->nr, RISCMEM_LEFT(ro, hw->risc_odd, hw->risc_len), RISCMEM_LEFT(re, hw->risc_even, hw->risc_len));
	*(ro++) = cpu_to_le32(BT848_RISC_JUMP);
	*(ro++) = cpu_to_le32(btv->cap_odd_end);
	*(re++) = cpu_to_le32(BT848_RISC_JUMP);
	*(re++) = cpu_to_le32(btv->cap_even_end);

	return 0;
}

static void bt848_stop_dma(struct bttv *btv)
{
	Dprintk(2,"bttv%d: Stopping DMA.\n", btv->nr);
	btv->dma = 0;
	btand(~15, BT848_GPIO_DMA_CTL);
}

static int bt848_set_dma(struct bttv *btv)
{
	struct bttvo *btvo;
	u32 *p;
	int i;
	
	btv->dma = 0; // make sure DMA is off for now
	btand(~15, BT848_GPIO_DMA_CTL);
	if(!btv->risc)
		btv->risc = (u32 *)kmalloc(PAGE_SIZE, GFP_KERNEL);
	if(!btv->risc)
	{
		printk("bttv%d: Could not malloc memory for risc table.\n", btv->nr);
		return -1;
	}
	if(!btv->vbi_dummy)
		btv->vbi_dummy = (u32 *)kmalloc(PAGE_SIZE, GFP_KERNEL);
	if(!btv->vbi_dummy)
	{
		printk("bttv%d: Could not malloc memory vbi dummy.\n", btv->nr);
		return -1;
	}
	btv->vbi_dummy_phys = cpu_to_le32(virt_to_bus(btv->vbi_dummy));

	/* Create basic table */
	p = btv->risc;
	*(p++) = cpu_to_le32(BT848_RISC_SYNC|BT848_RISC_RESYNC|BT848_FIFO_STATUS_VRE);
	*(p++) = cpu_to_le32(0);
	*(p++) = cpu_to_le32(BT848_RISC_JUMP);
	btv->vbi_odd_jmp = p++;
	btv->vbi_odd_end = cpu_to_le32(virt_to_bus(p));
	*(p++) = cpu_to_le32(BT848_RISC_JUMP);
	btv->cap_odd_jmp = p++;
	btv->cap_odd_end = cpu_to_le32(virt_to_bus(p));
	*(p++) = cpu_to_le32(BT848_RISC_JUMP|BT848_RISC_IRQ|(0x01<<20));
	*p = cpu_to_le32(virt_to_bus(p+1)); /* just jump one ahead */
	p++;

	*(p++) = cpu_to_le32(BT848_RISC_SYNC|BT848_RISC_RESYNC|BT848_FIFO_STATUS_VRO);
	*(p++) = cpu_to_le32(0);
	*(p++) = cpu_to_le32(BT848_RISC_JUMP);
	btv->vbi_even_jmp = p++;
	btv->vbi_even_end = cpu_to_le32(virt_to_bus(p));
	*(p++) = cpu_to_le32(BT848_RISC_JUMP);
	btv->cap_even_jmp = p++;
	btv->cap_even_end = cpu_to_le32(virt_to_bus(p));
	*(p++) = cpu_to_le32(BT848_RISC_JUMP|BT848_RISC_IRQ|(0x01<<16));
	*(p++) = cpu_to_le32(virt_to_bus(btv->risc)); /* just to start */
	
	/* vbi jump tables */
	
	btv->vbi_odd_start = cpu_to_le32(virt_to_bus(p));
	*(p++) = cpu_to_le32(BT848_RISC_SYNC|BT848_FIFO_STATUS_FM1);
	*(p++) = 0;
	for(i = 0; i < (tvnorms[btv->norm].vdelay/2); i++)
	{
		*(p++) = cpu_to_le32(BT848_RISC_WRITE|BT848_RISC_EOL|BT848_RISC_SOL|btv->vbi_bpl);
		if(i==0)
			btv->vbi_odd_tab = p;
		*(p++) = btv->vbi_dummy_phys;
	}
	*(p++) = cpu_to_le32(BT848_RISC_JUMP);
	*(p++) = btv->vbi_odd_end;

	btv->vbi_even_start = cpu_to_le32(virt_to_bus(p));
	*(p++) = cpu_to_le32(BT848_RISC_SYNC|BT848_FIFO_STATUS_FM1);
	*(p++) = 0;
	for(i = 0; i < (tvnorms[btv->norm].vdelay/2); i++)
	{
		*(p++) = cpu_to_le32(BT848_RISC_WRITE|BT848_RISC_EOL|BT848_RISC_SOL|btv->vbi_bpl);
		if(i==0)
			btv->vbi_even_tab = p;
		*(p++) = btv->vbi_dummy_phys;
	}
	*(p++) = cpu_to_le32(BT848_RISC_JUMP);
	*(p++) = btv->vbi_even_end;

	
	/* fill in cap dummy */
	if(!btv->dummy_buf.dma)
		btv->dummy_buf.dma = (struct scatter_node *)kmalloc(PAGE_SIZE, GFP_KERNEL);
	if(!btv->dummy_buf.dma)
	{
		printk("bttv%d: failed to allocate memory for default dma list - aborting.\n", btv->nr);
		return -1;
	}
#if 1
	Dprintk(2,"bttv%d: using dummy to skip\n", btv->nr);
	btv->dummy_buf.dma->addr= 0;
	btv->dummy_buf.dma->len = END_OF_SCATTER_LIST|0x7fffffffUL;
	bt848_calc_geo(btv, &btv->dummy_geo, btv->dummy_width, btv->dummy_height, V4L2_PIX_FMT_RGB565);
	if(make_risctab(btv, &btv->dummy_buf, btv->dummy_width, btv->dummy_height, V4L2_PIX_FMT_RGB565, 3200, NULL)!=0)
	{
		printk("bttv%d: failed to build dummy risc table - aborting.\n", btv->nr);
		return -1;
	}
#else
	Dprintk(2,"bttv%d: using dummy to FB\n", btv->nr);
	btv->dummy_buf.dma->addr=0xd6000000UL;
	btv->dummy_buf.dma->len = END_OF_SCATTER_LIST|0x7fffffffUL;
	bt848_calc_geo(btv, &btv->dummy_geo, btv->dummy_width, btv->dummy_height, V4L2_PIX_FMT_RGB565);
	if(make_risctab(btv, &btv->dummy_buf, btv->dummy_width, btv->dummy_height, V4L2_PIX_FMT_RGB565, 3200, NULL)!=0)
	{
		printk("bttv%d: failed to build dummy risc table - aborting.\n", btv->nr);
		return -1;
	}
#endif
	*btv->cap_odd_jmp = btv->dummy_buf.risc_odd_bus;
	*btv->cap_even_jmp = btv->dummy_buf.risc_even_bus;
	bt848_set_geo(btv, &btv->dummy_geo, 0);
	bt848_set_geo(btv, &btv->dummy_geo, 1);
	btv->e_stat = STAT_DUMMY;
	btv->o_stat = STAT_DUMMY;
	for(i=0; i<(MAX_OPENS+MAX_VBI_OPENS); i++)
	{
		btvo = (i>=MAX_OPENS)?&btv->vbi_open_data[i-MAX_OPENS]:&btv->open_data[i];
		if(!btvo->isopen)
			continue;
		if(btvo->read.wait_even)
		{
			btvo->read.wait_even->buf.flags &= ~V4L2_BUF_FLAG_QUEUED;
			btvo->read.wait_even = NULL;
		}
		if(btvo->read.wait_odd)
		{
			btvo->read.wait_odd->buf.flags &= ~V4L2_BUF_FLAG_QUEUED;
			btvo->read.wait_odd = NULL;
		}
		if(btvo->stream.wait_even)
		{
			btvo->stream.wait_even->buf.flags &= ~V4L2_BUF_FLAG_QUEUED;
			btvo->stream.wait_even = NULL;
		}
		if(btvo->stream.wait_odd)
		{
			btvo->stream.wait_odd->buf.flags &= ~V4L2_BUF_FLAG_QUEUED;
			btvo->stream.wait_odd = NULL;
		}
	}
	btwrite(cpu_to_le32(virt_to_bus(btv->risc)), BT848_RISC_STRT_ADD);
	btand(~15, BT848_CAP_CTL);
	if(btv->vbi_opens>0)
	{
		btor(15, BT848_CAP_CTL);
		*btv->vbi_odd_jmp = btv->vbi_odd_start;
		*btv->vbi_even_jmp = btv->vbi_even_start;
	} else
	{
		btor(3, BT848_CAP_CTL);
		*btv->vbi_odd_jmp = btv->vbi_odd_end;
		*btv->vbi_even_jmp = btv->vbi_even_end;
	}
	btor(15, BT848_GPIO_DMA_CTL);
	btv->dma = 1;

	return 0;
}

static void bt848_open(struct bttv *btv, int vbi)
{
	if(vbi)
	{
		++btv->vbi_opens;
		if(btv->vbi_opens==1) /* First vbi open */
		{
			/* Bt848 will automatically skip unhandled vbi data, so 
			   we start the captuee first, then the RISC. As 
			   long as these two writes are atomic 32 bit operations
			   everything should be safe */
			btor(BT848_CAP_CTL_CAPTURE_VBI_ODD|BT848_CAP_CTL_CAPTURE_VBI_EVEN, BT848_CAP_CTL);
			*btv->vbi_odd_jmp = btv->vbi_odd_start;
			*btv->vbi_even_jmp = btv->vbi_even_start;
			Dprintk(2, "bttv%d: vbi capture started.\n", btv->nr);
		}
	}
	++btv->open_count;
	if(btv->open_count==1)
	{
		/* Make sure preview and capture are turned off, and fire up the dma engine */
		btv->pv=0;
		btv->streaming_caps = NULL;
		bt848_set_dma(btv);
		Dprintk(2, "bttv%d: video capture started.\n", btv->nr);
	}
}

static void bt848_close(struct bttv *btv, int vbi)
{
	if(vbi)
	{
		--btv->vbi_opens;
		if(btv->vbi_opens==0)
		{
			/* Bt848 will automatically skip unhandled vbi data, so 
			   we stop the RISC first, then turn off capture.  As 
			   long as these two writes are atomic 32 bit operations
			   everything should be safe */
			*btv->vbi_odd_jmp = btv->vbi_odd_end;
			*btv->vbi_even_jmp = btv->vbi_even_end;
			btand(~(BT848_CAP_CTL_CAPTURE_VBI_ODD|BT848_CAP_CTL_CAPTURE_VBI_EVEN), BT848_CAP_CTL);
			Dprintk(2, "bttv%d: vbi capture stopped.\n", btv->nr);
		}
	}
	--btv->open_count;
	if(btv->open_count==0)
	{
		/* Make sure preview and capture are turned off, and fire up the dma engine */
		bt848_stop_dma(btv);
		Dprintk(2, "bttv%d: video capture stopped.\n", btv->nr);
	}
}

/* Only called by module remove - no need to clean up pointers */

static void bt848_free_dma(struct bttv *btv)
{
	Dprintk(2,"bttv%d: Freeing dma structures.\n", btv->nr);
	if(btv->dma)
		bt848_stop_dma(btv);
	if(btv->risc)
	{
		kfree(btv->risc);
		btv->risc = NULL;
	}
	if(btv->vbi_dummy)
	{
		kfree(btv->vbi_dummy);
		btv->vbi_dummy = NULL;
	}
	if(btv->dummy_buf.risc_even)
	{
		kfree(btv->dummy_buf.risc_even);
		btv->dummy_buf.risc_even = NULL;
	}
	if(btv->dummy_buf.risc_odd)
	{
		kfree(btv->dummy_buf.risc_odd);
		btv->dummy_buf.risc_odd = NULL;
	}
	if(btv->dummy_buf.dma)
	{
		kfree(btv->dummy_buf.dma);
		btv->dummy_buf.dma = NULL;
	}
}

static long dma_sync(struct bttv *btv)
{
	long to;
	unsigned long flags;
	DECLARE_WAITQUEUE(wait, current);
	
	to = HZ/10; /* at least two field periods */
	
	spin_lock_irqsave(&btv->s_lock, flags);
	btv->sync++; /* request sync */
	spin_unlock_irqrestore(&btv->s_lock, flags);
	add_wait_queue(&btv->riscq, &wait);
	current->state = TASK_INTERRUPTIBLE;
	while((to!=0)&&(btv->sync!=0))
		to = schedule_timeout(to);
	remove_wait_queue(&btv->riscq, &wait);
	current->state = TASK_RUNNING;
	if(to==0)
		Dprintk(1,"bttv%d: Timeout waiting for irq!\n", btv->nr);
	return to;
}

/* PREVIEW FUNCTIONS */

/* Only caalled in module remove */

static void free_preview(struct bttv *btv)
{
	if(btv->win.bitmap)
	{
		kfree(btv->win.bitmap);
		btv->win.bitmap = NULL;
	}
	if(btv->pv_buf.risc_even)
	{	
		kfree(btv->pv_buf.risc_even);
		btv->pv_buf.risc_even = NULL;
	}
	if(btv->pv_buf.risc_odd)
	{
		kfree(btv->pv_buf.risc_odd);
		btv->pv_buf.risc_odd = NULL;
	}
	if(btv->pv_buf.dma)
	{
		kfree(btv->pv_buf.dma);
		btv->pv_buf.dma = NULL;
	}
}

static void clipmap_rectangle(struct bttv *btv, struct v4l2_clip *clip)
{
	int j;
	u32 s_ofs, e_ofs;
	u8 mask;
	
	Dprintk(2,"bttv%d: clip %d %d %d %d\n", btv->nr, (*clip).x, (*clip).y, (*clip).width, (*clip).height);
	if((*clip).x<0)
	{
		(*clip).width+=(*clip).x;
		(*clip).x=0;
	}
	if((*clip).y<0)
	{
		(*clip).height+=(*clip).y;
		(*clip).y=0;
	}
	if((*clip).x+(*clip).width > btv->win.width)
		(*clip).width = btv->win.width - (*clip).x;
	if((*clip).y+(*clip).height > btv->win.height)
		(*clip).height = btv->win.height - (*clip).y;
	if(((*clip).x>=btv->win.width)
	   ||((*clip).y>=btv->win.height)
	   ||((*clip).width<=0)
	   ||((*clip).height<=0))
	{
		Dprintk(1,"bttv%d: Whoa - bad clip -> ignoring!\n", btv->nr);
		return;
	}
	for(j=(*clip).y; j<(*clip).y+(*clip).height; j++)
	{
		s_ofs = (j*btv->win.width)+(*clip).x;
		e_ofs = s_ofs + (*clip).width;
		if((s_ofs/8)==(e_ofs/8))
		{
			for(mask = 0x80>>(s_ofs%8); mask != 0x80>>(e_ofs%8); mask>>=1)
				((u8*)btv->win.bitmap)[s_ofs/8]|=mask;
			continue;
		}
		mask = 0x80>>(s_ofs%8);
		while((s_ofs%8)!=0)
		{
			((u8*)btv->win.bitmap)[s_ofs/8]|=mask;
			s_ofs++;
			mask>>=1;
		}
		while((s_ofs/8)!=(e_ofs/8))
		{
			((u8*)btv->win.bitmap)[s_ofs/8]|=0xff;
			s_ofs+=8;
		}
		mask = 0x80;
		while((s_ofs%8)!=(e_ofs%8))
		{
			((u8*)btv->win.bitmap)[s_ofs/8]|=mask;
			s_ofs++;
			mask>>=1;
		}
	}
}

static int pv_new_win(struct bttv *btv, void *cm)
{
 	int skip, width, height, i;
	u32 bmsize;
	struct v4l2_clip clip;
	
	width = btv->win.width;
	height = btv->win.height;

	if(btv->pv)
	{
		printk("bttv%d: cannot build preview risc table while preview is active - aborting.\n", btv->nr);
		return -EINVAL;
	}

	if(btv->win.width > tvnorms[btv->norm].maxwidth)
		btv->win.width = tvnorms[btv->norm].maxwidth;
	if(btv->win.height > tvnorms[btv->norm].sheight)
		btv->win.height = tvnorms[btv->norm].sheight;
	if(btv->win.width < 16)
		btv->win.width = 16;
	if(btv->win.height < 16)
		btv->win.height = 16;
	btv->win.width = ((btv->win.width + 15)>>4)<<4;
	btv->win.height = ((btv->win.height + 15)>>4)<<4;
	
	bmsize = ((btv->win.width*btv->win.height)+7)>>3;
	if((btv->win.bitmap)&&(btv->bmsize<bmsize))
	{
		kfree(btv->win.bitmap);
		btv->win.bitmap = NULL;
	}
	if(!btv->win.bitmap)
		btv->win.bitmap=kmalloc(bmsize, GFP_KERNEL);
	if(!btv->win.bitmap)
	{
		printk("bttv%d: Could not allocate memory for bitmap.\n", btv->nr);
		btv->bmsize = 0;
		return -ENOMEM;
	}
	btv->bmsize = bmsize;
	if((btv->win.bitmap)&&(width==btv->win.width)&&(height==btv->win.height)&&(btv->win.clipcount<0))
	{
		if(copy_from_user(btv->win.bitmap, cm, bmsize))
			return -EFAULT;
	}
	else
	{
		memset(btv->win.bitmap, 0, bmsize);
	}
	
	for(i = 0; i<btv->win.clipcount; i++)
	{
		if(copy_from_user(&clip, &btv->win.clips[i], sizeof(struct v4l2_clip)))
			return -EFAULT;
		/* clip.x = btv->win.clips[i].x;
		clip.y = btv->win.clips[i].y;
		clip.width = btv->win.clips[i].width;
		clip.height = btv->win.clips[i].height; */
		clipmap_rectangle(btv, &clip);
	}
	if(btv->win.x<0)
	{
		clip.x = 0;
		clip.y = 0;
		clip.height = btv->win.height;
		clip.width = - btv->win.x;
		clipmap_rectangle(btv, &clip);
	}
	if(btv->win.y < 0)
	{
		clip.x = 0;
		clip.y = 0;
		clip.width = btv->win.width;
		clip.height = - btv->win.y;
		clipmap_rectangle(btv, &clip);
	}
	if(btv->win.x + btv->win.width > btv->fb.fmt.width)
	{
		clip.y = 0;
		clip.height = btv->win.height;
		clip.width = (btv->win.x + btv->win.width) - btv->fb.fmt.width;
		clip.x = btv->win.width - clip.width;
		clipmap_rectangle(btv, &clip);
	}
	if(btv->win.y + btv->win.height > btv->fb.fmt.height)
	{
		clip.x = 0;
		clip.width = btv->win.width;
		clip.height = (btv->win.y + btv->win.height) - btv->fb.fmt.height;
		clip.y = btv->win.height - clip.height;
		clipmap_rectangle(btv, &clip);
	}
	
	skip = (btv->win.y*btv->fb.fmt.bytesperline) + ((btv->win.x*btv->fb.fmt.depth + 7)>>3);
	
	if(!btv->pv_buf.dma)
	btv->pv_buf.dma = (struct scatter_node *)kmalloc(PAGE_SIZE, GFP_KERNEL);
	if(!btv->pv_buf.dma)
	{
		printk("bttv%d: failed to allocate memory for preview dma list - aborting.\n", btv->nr);
		return -1;
	}
	btv->pv_buf.dma->addr=(__u32)btv->fb.base[0] + skip;
	btv->pv_buf.dma->len = END_OF_SCATTER_LIST|(((btv->fb.fmt.width * btv->fb.fmt.height * btv->fb.fmt.depth + 7)>>3) - skip);
	bt848_calc_geo(btv, &btv->pv_geo, btv->win.width, btv->win.height, btv->fb.fmt.pixelformat);
	if(make_risctab(btv, &btv->pv_buf, btv->win.width, btv->win.height, btv->fb.fmt.pixelformat, btv->fb.fmt.bytesperline, btv->win.bitmap)!=0)
	{
		printk("bttv%d: failed to build preview risc table - aborting.\n", btv->nr);
		return -EINVAL;
	}
	Dprintk(2,"bttv%d: pv_new_win %dx%d @ %d %d\n", btv->nr, btv->win.width, btv->win.height, btv->fb.fmt.depth, btv->fb.fmt.bytesperline);
	return 0;
}

static int preview(struct bttv *btv, int onoff)
{
	unsigned long flags;
	
	if((btv->pvstat!=PV_WIN)&&(btv->pvstat!=PV_OVL))
	{
		Dprintk(1,"bttv%d: Cannot activate preview without window/overlay.\n", btv->nr);
		return -EINVAL;
	}
	if(onoff)
	{
		Dprintk(1,"bttv%d: Turning preview on.\n", btv->nr);
		if((btv->input!=btv->pv_open->input)&&
		   (!btv->streaming_caps)&&
		   (!btv->streaming_vbis))
			bt848_set_input(btv, btv->pv_open->input);
		spin_lock_irqsave(&btv->s_lock, flags); 
		btv->pv=1;
		spin_unlock_irqrestore(&btv->s_lock, flags);
	} else
	{
		int i = 0;
		int fst;
		
		Dprintk(1,"bttv%d: Turning preview off.\n", btv->nr);
		spin_lock_irqsave(&btv->s_lock, flags); 
		btv->pv=0;
		spin_unlock_irqrestore(&btv->s_lock, flags);
		do
		{
			if((i++>2)||(dma_sync(btv)==0))
			{
				Dprintk(2,"bttv%d: RISC IRQ not responding - resetting DMA engine\n", btv->nr);
				bt848_stop_dma(btv);
				bt848_set_dma(btv);
				break;
			}
			spin_lock_irqsave(&btv->s_lock, flags); 
			fst = ((btv->e_stat==STAT_PV)||(btv->o_stat==STAT_PV));
			spin_unlock_irqrestore(&btv->s_lock, flags);
		} while(fst);
	}
	return 0;
}

/***********************************/
/* Capture functions               */
/***********************************/

static int capture_set_format(struct bttvo *btvo, struct v4l2_pix_format *f)
{
	struct bttv *btv = btvo->dev;
	int i;
	int maxwidth, maxheight;
	
	Dprintk(2,"bttv%d: cap_set_format %dx%d for %d\n", btv->nr, f->width, f->height, btvo->index);
	
	i=find_format(f->pixelformat);
	if(i<0)
	{
		Dprintk(1,"bttv%d: s_fmt - invalid pixel format\n", btv->nr);
		return -EINVAL;
	}
	f->depth = formats[i].depth;
	maxwidth = tvnorms[btv->norm].maxwidth;
	maxheight = tvnorms[btv->norm].sheight;
	if((f->flags&V4L2_FMT_FLAG_INTERLACED)==0) // interlace not allowed
		maxheight/=2;
	if(f->width>maxwidth)
		f->width = maxwidth;
	if(f->height>maxheight)
		f->height = tvnorms[btv->norm].sheight;
	f->width&=~15;
	f->height&=~15;
	if(f->width<16)
		f->width = 16;
	if(f->height<16)
		f->height = 16;
	f->sizeimage = (f->width * f->height * f->depth + 7)>>3;
	if(btvo->stream.map)
	{
		if(f->sizeimage > btvo->cap_bsize)
		{
			Dprintk(1,"bttv%d: s_fmt - invalid image size\n", btv->nr);
			return -EINVAL;
		}
	} else
	{
		btvo->cap_bsize = (f->sizeimage + PAGE_SIZE - 1) & ~(PAGE_SIZE-1);
	}
	if(f->height>(tvnorms[btv->norm].sheight/2))
	{
		f->flags|=V4L2_FMT_FLAG_INTERLACED;
		f->flags&=~(V4L2_FMT_FLAG_TOPFIELD|V4L2_FMT_FLAG_BOTFIELD);
	}
	else
		f->flags&=~V4L2_FMT_FLAG_INTERLACED;
	if((f->flags&V4L2_FMT_FLAG_BYTESPERLINE)==0)
		f->bytesperline = (f->width * f->depth + 7)>>3;
	memcpy(&btvo->cap_fmt, f, sizeof(struct v4l2_pix_format));
	if(!btvo->cap_dummy.dma)
		btvo->cap_dummy.dma = (struct scatter_node *)kmalloc(PAGE_SIZE, GFP_KERNEL);
	if(!btvo->cap_dummy.dma)
	{
		printk("bttv%d: failed to allocate memory for capture dummy dma list - aborting.\n", btv->nr);
		return -1;
	}
	btvo->cap_dummy.dma->addr= 0;
	btvo->cap_dummy.dma->len = END_OF_SCATTER_LIST|0x7fffffffUL;
	bt848_calc_geo(btv, &btvo->cap_geo, f->width, f->height, f->pixelformat);
	if(make_risctab(btv, &btvo->cap_dummy, f->width, f->height, f->pixelformat, 0, NULL)!=0)
	{
		 printk("bttv%d: failed to build cap_dummy risc table - aborting.\n", btv->nr);
		 return -EINVAL;
	}
	memcpy(&btv->cap_fmt, f, sizeof(struct v4l2_pix_format));
	return 0;
}

static int capture_request_buffers(struct bttvc *btvc, struct v4l2_requestbuffers *req)
{
	struct bttvo *btvo = btvc->open;
	struct bttv *btv = btvo->dev;
	int max_count, i;
	
	if(btvc->req)
	{
		Dprintk(1,"bttv%d: buffers already requested!\n", btv->nr);
		return -EINVAL;
	}
	req->type=V4L2_BUF_TYPE_CAPTURE;
	btvc->cap_contig = 0;
	if(req->type & V4L2_BUF_REQ_CONTIG)
	{
		btvc->cap_contig = 1;
		req->type |= V4L2_BUF_REQ_CONTIG;
	}
	if(req->count < 1)
		req->count = 1;
	/* FIXME calculate based on other opens too */
	max_count = (bttv_maxmem * 1024 * 1024) / btvo->cap_bsize;
	if(req->count > max_count)
		req->count = max_count;
	Dprintk(1,"bttv%d: Granting %d buffers.\n", btv->nr, req->count);
	btvc->bufs = vmalloc_32(sizeof(struct stream_buffer)*req->count);
	if(!btvc->bufs)
	{
		printk("bttv%d: Could not vmalloc() capture buffers.\n", btv->nr);
		return -ENOMEM;
	}
	btvc->buf_vmalloc = sizeof(struct stream_buffer)*req->count;
	memset(btvc->bufs, 0, btvc->buf_vmalloc);
	for(i=0; i<req->count; i++)
	{
		btvc->bufs[i].buf.index = i;
		btvc->bufs[i].buf.type = req->type;
		btvc->bufs[i].buf.offset = i*btvo->cap_bsize;
		btvc->bufs[i].buf.length = btvo->cap_bsize;
		btvc->bufs[i].c = btvc;
	}
	btvc->req = req->count;
	
	return 0;
}

static void capture_unrequest_buffers(struct bttvc *btvc)
{
	struct bttvo *btvo = btvc->open;
	struct bttv *btv = (struct bttv*)btvo->dev;
	int i;
	
	if(!btvc->req)
	{
		Dprintk(2,"bttv%d: caputre buffers notrequested!\n", btv->nr);
		return;
	}

	while (v4l2_q_del_head(&btvc->cap_done_q));
	while (v4l2_q_del_head(&btvc->cap_wait_q));
	
	if(!btvc->bufs)
	{
		printk("bttv%d: Hello... No buffers!\n", btv->nr);
		return;
	}

	for(i=0; i<btvc->req; i++)
	{
		if(btvc->bufs[i].hw.dma)
			kfree(btvc->bufs[i].hw.dma);
		if(btvc->bufs[i].hw.risc_even)
			kfree(btvc->bufs[i].hw.risc_even);
		if(btvc->bufs[i].hw.risc_odd)
			kfree(btvc->bufs[i].hw.risc_odd);
	}
	if(btvc->bufs[0].hw.data)
		rvfree(btvc->bufs[0].hw.data, btvc->cap_vmalloc);
	vfree(btvc->bufs);
	btvc->bufs = NULL;
	btvc->buf_vmalloc = 0;
	btvc->cap_vmalloc = 0;
	btvc->req = 0;
	
	return;
}

static int capture_queuebuffer(struct bttvc *btvc, struct v4l2_buffer *vbuf)
{
	struct bttvo *btvo = btvc->open;
	struct bttv *btv = btvo->dev;
	int i = vbuf->index;
	struct stream_buffer *buf = NULL;

	if (vbuf->type != V4L2_BUF_TYPE_CAPTURE)
	{
		Dprintk(1,"bttv%d: QBUF wrong type\n", btv->nr);
		return -EINVAL;
	}
	if ((i < 0) || (i >= btvc->req))
	{
		Dprintk(1,"bttv%d: QBUF buffer index %d is out of range\n", btv->nr, i);
		return -EINVAL;
	}
	buf = &btvc->bufs[i];
	if ((buf->buf.flags & V4L2_BUF_FLAG_QUEUED))
	{
		Dprintk(1,"bttv%d: QBUF buffer %d is already queued\n", btv->nr, i);
		return -EINVAL;
	}
	buf->buf.flags &= ~V4L2_BUF_FLAG_DONE;
	buf->buf.flags |= V4L2_BUF_FLAG_QUEUED;
	v4l2_q_add_tail(&btvc->cap_wait_q, &buf->q);
	return 0;
}

static int capture_dequeuebuffer(struct bttvc *btvc, struct v4l2_buffer *buf)
{
	struct bttvo *btvo = btvc->open;
	struct bttv *btv = (struct bttv*)btvo->dev;
	static struct stream_buffer *newbuf;
	int odd;

	if (buf->type != V4L2_BUF_TYPE_CAPTURE)
	{
		Dprintk(1,"bttv%d: DQBUF wrong buffer type\n", btv->nr);
		return -EINVAL;
	}
	newbuf = v4l2_q_del_head(&btvc->cap_done_q);
	if (newbuf == NULL)
	{
		Dprintk(1,"bttv%d: DQBUF nothing on done queue\n", btv->nr);
		return -EINVAL;
	}
	
	odd = newbuf->buf.bytesused;
	if(btvc ->start_time == 0)
		btvc->start_time = newbuf->buf.timestamp;
	btvc->int_time = newbuf->buf.timestamp;
	btvc->int_time -= btvc->start_time;
	btvo->perf.frames++;
	btvo->perf.bytesout += btvo->cap_fmt.sizeimage;
	newbuf->buf.bytesused = btvo->cap_fmt.sizeimage;
	newbuf->buf.flags |= V4L2_BUF_FLAG_DONE | V4L2_BUF_FLAG_KEYFRAME;
	newbuf->buf.flags &= ~(V4L2_BUF_FLAG_BOTFIELD|V4L2_BUF_FLAG_TOPFIELD);
	if((btvo->cap_fmt.flags&(V4L2_FMT_FLAG_TOPFIELD|V4L2_FMT_FLAG_BOTFIELD))
	   == (V4L2_FMT_FLAG_TOPFIELD|V4L2_FMT_FLAG_BOTFIELD))
		newbuf->buf.flags |= (odd)?V4L2_BUF_FLAG_TOPFIELD:V4L2_BUF_FLAG_BOTFIELD;
	newbuf->buf.timestamp = btvc->int_time;
	if(btvc->last_frame>=0)
	{
		btvc->frame_no = v4l2_timestamp_correct(&newbuf->buf.timestamp, btv->tpf);
		newbuf->buf.timestamp += btvc->start_time;
		if(btvc->frame_no > btvc->last_frame+1)
		btvo->perf.framesdropped += btvc->frame_no - (btvc->last_frame + 1);
	} else
	{
		btvc->frame_no = 0;
		newbuf->buf.timestamp = btvc->start_time;
	}
	newbuf->buf.sequence = btvc->frame_no;
	btvc->last_frame = btvc->frame_no;
	
	newbuf->buf.flags &= ~V4L2_BUF_FLAG_QUEUED;
	memcpy(buf, &newbuf->buf, sizeof(struct v4l2_buffer));
	return 0;
}
 
/*
   When this returns the open has NO dma resources anywhere, 
   including wait_even and wait_odd
*/
 
static void capture_streamoff(struct bttvc *btvc, __u32 type)
{
	struct bttvo *btvo = btvc->open;
	struct bttv *btv = btvo->dev;
	struct bttvc *c;
	unsigned long flags;
	int i = 0;
	int fst;
	
	Dprintk(2,"bttv%d: streamoff\n", btv->nr);
	for(c=btv->streaming_caps; c; c=c->next)
		if(c==btvc)
			break;
	if(!c)
		return;
	spin_lock_irqsave(&btv->s_lock, flags); 
	btvc->enable = 0;
	spin_unlock_irqrestore(&btv->s_lock, flags);
	do
	{
		if((i++>3)||(dma_sync(btv)==0))
		{
			Dprintk(2,"bttv%d: RISC IRQ not responding - resetting DMA engine\n", btv->nr);
			bt848_stop_dma(btv);
			bt848_set_dma(btv);
			break;
		}
		spin_lock_irqsave(&btv->s_lock, flags); 
		fst = ((btvc->wait_even)||(btvc->wait_odd));
		spin_unlock_irqrestore(&btv->s_lock, flags);
	} while(fst);
	spin_lock_irqsave(&btv->s_lock, flags);
	if(btv->streaming_caps==btvc)
	{
		btv->streaming_caps = btvc->next;
	} else
	{
		for(c=btv->streaming_caps; c->next; c = c->next)
		{
			if(c->next==btvc) c->next = btvc->next;
			break;
		}
	}
	spin_unlock_irqrestore(&btv->s_lock, flags);
	if((btv->pv_open)&&
	   (btv->pv_open->input!=btv->input)&&
	   (!btv->streaming_caps)&&
	   (!btv->streaming_vbis))
		bt848_set_input(btv, btv->pv_open->input);
	Dprintk(2,"bttv%d: streamoff done\n", btv->nr);
	return;
}

static int capture_streamon(struct bttvc *btvc, __u32 type)
{
	struct bttvo *btvo = btvc->open;
	struct bttv *btv = btvo->dev;
	struct bttvc *c;
	struct stream_buffer *buf;
	int i, even=0, odd=0;
	unsigned long flags;
	
	Dprintk(2,"bttv%d: streamon\n", btv->nr);
	if (type != V4L2_BUF_TYPE_CAPTURE)
	{
		Dprintk(1,"bttv%d: wrong buffer type!\n", btv->nr);
		return -EINVAL;
	}
	if(btvo->input!=btv->input)
	{
		if((btv->streaming_caps)||(btv->streaming_vbis))
		{
			Dprintk(1,"bttv%d: input is busy\n", btv->nr);
			return -EBUSY;
		}
		bt848_set_input(btv, btvo->input);
	}
	for(c=btv->streaming_caps; c; c=c->next)
	{
		if(c==btvc)
		{
			Dprintk(2,"bttv%d: streaming already on!\n", btv->nr);
			return -EBUSY;
		}
		if(c->open->cap_fmt.flags & V4L2_FMT_FLAG_INTERLACED)
		{
			even++;
			odd++;
		} else if(c->open->cap_fmt.flags & V4L2_FMT_FLAG_TOPFIELD)
		{
			odd++;
		} else if(c->open->cap_fmt.flags & V4L2_FMT_FLAG_BOTFIELD)
		{
			even++;
		} 
	}
	if( (btvo->cap_fmt.flags&V4L2_FMT_FLAG_INTERLACED)&&
	    (even|odd) )
	{
		Dprintk(2,"bttv%d: interlace requested while streaming\n", btv->nr);
		return -EBUSY;
	}
	if((btvo->cap_fmt.flags&V4L2_FMT_FLAG_TOPFIELD)&&(odd))
	{
		Dprintk(2,"bttv%d: odd requested while streaming\n", btv->nr);
		return -EBUSY;
	}
	if((btvo->cap_fmt.flags&V4L2_FMT_FLAG_BOTFIELD)&&(even))
	{
		Dprintk(2,"bttv%d: even requested while streaming\n", btv->nr);
		return -EBUSY;
	}
	if((even)&&(odd))
	{
		Dprintk(2,"bttv%d: single requested while both busy\n", btv->nr);
		return -EBUSY;
	}
	if(!(btvo->cap_fmt.flags&V4L2_FMT_FLAG_FIELD_field))
	{
		if(even)
			btvo->cap_fmt.flags|=V4L2_FMT_FLAG_TOPFIELD;
		else
			btvo->cap_fmt.flags|=V4L2_FMT_FLAG_BOTFIELD;
	}
	while ((buf = v4l2_q_del_head(&btvc->cap_done_q)))
		buf->buf.flags &= ~V4L2_BUF_FLAG_QUEUED;
	for(i=0; i<btvc->req; i++)
	{
		if(!btvc->bufs[i].hw.dma)
			btvc->bufs[i].hw.dma=(struct scatter_node *)kmalloc(PAGE_SIZE, GFP_KERNEL);
		if(!btvc->bufs[i].hw.dma)
		{
			printk("bttv%d: could not kmalloc() dma tables!\n", btv->nr);
			return -ENOMEM;
		}
		if(bm_build_scatter_list(btvc, btvc->bufs[i].hw.data, &btvc->bufs[i].hw.dma)<0)
		{
			printk("bttv%d: could not build scatter list!\n", btv->nr);
			return -ENOMEM;
		}
		if(make_risctab(btv, &btvc->bufs[i].hw, btvo->cap_fmt.width, btvo->cap_fmt.height, btvo->cap_fmt.pixelformat, btvo->cap_fmt.bytesperline, NULL)!=0)
		{
			printk("bttv%d: could not build risc tables!\n", btv->nr);
			return -EINVAL;
		}
	}
	memset(&btvo->perf, 0, sizeof(struct v4l2_performance));
	btvc->start_time = 0;
	btvc->last_frame = -1;
	btvc->wait_even = NULL;
	btvc->wait_odd = NULL;
	
	btvc->enable = 1;
	spin_lock_irqsave(&btv->s_lock, flags);
	btvc->next = btv->streaming_caps;
	btv->streaming_caps = btvc;
	spin_unlock_irqrestore(&btv->s_lock, flags);
	
	Dprintk(2,"bttv%d: streamon done\n", btv->nr);
	return 0;
}


/***********************************/
/* MMAP functions                  */
/***********************************/

static struct stream_buffer * mmap_stream_buffer_from_offset(struct bttvc *btvc, unsigned long offset)
{
	int	i;
	for (i = 0; i < btvc->req; i++)
		if (offset == btvc->bufs[i].buf.offset)
			return &btvc->bufs[i];
	return NULL;
}

#ifndef BTTV_MM

static void mmap_vma_open(struct vm_area_struct *vma)
{
	struct bttv *btv = (struct bttv*)v4l2_device_from_file(vma->vm_file);
	struct bttvo *btvo = (struct bttvo*)v4l2_openid_from_file(vma->vm_file);
	struct bttvc *btvc = &btvo->stream;
 	struct stream_buffer *buf;

	if ((!btv)||(!btvo))
		return;
		
	if (btvc->cap_contig)
		buf = &btvc->bufs[0];
	else
#if  LINUX_VERSION_CODE > KERNEL_VERSION(2,3,0)
		buf = mmap_stream_buffer_from_offset(btvc, vma->vm_pgoff * PAGE_SIZE);
#else
		buf = mmap_stream_buffer_from_offset(btvc, vma->vm_offset);
#endif
	buf->hw.vma_nopen++;
	Dprintk(2,"bttv%d: vma_open called on buffer %d (refcount %d)\n",
		btv->nr, buf->buf.index, buf->hw.vma_nopen);
}

static void mmap_vma_close(struct vm_area_struct *vma)
{
	struct bttv *btv = (struct bttv*)v4l2_device_from_file(vma->vm_file);
	struct bttvo *btvo = (struct bttvo*)v4l2_openid_from_file(vma->vm_file);
	struct bttvc *btvc = &btvo->stream;
 	struct stream_buffer *buf;
	int i, n;
	struct bttvc *c;
	
	if((!btv)||(!btvo))
		return;
	if (btvc->cap_contig)
	{
		n = btvo->map;
		buf = &btvc->bufs[0];
	} else
	{
#if  LINUX_VERSION_CODE > KERNEL_VERSION(2,3,0)
		buf = mmap_stream_buffer_from_offset(btvc, vma->vm_pgoff * PAGE_SIZE);
#else
		buf = mmap_stream_buffer_from_offset(btvc, vma->vm_offset);
#endif
		n = 1;
	}
	if(!buf)
	{
		Dprintk(1,"bttv%d: munmap: invalid buffer!\n", btv->nr);
		return;
	}
	buf->hw.vma_nopen--;
	if (buf->hw.vma_nopen > 0)
		return;

	for(c=btv->streaming_caps; c; c=c->next)
		if(c==btvc)
		{
			Dprintk(1,"bttv%d: Warning - final munmap() called while streaming\n", btv->nr);
			capture_streamoff(btvc, V4L2_BUF_TYPE_CAPTURE);
		}

	for (i = 0; i < n; i++)
	{
		buf->buf.flags &= ~V4L2_BUF_FLAG_MAPPED;
		--btvc->map;
		Dprintk(1,"bttv%d: vma_close called on buffer %d (refcount %d)\n",

		btv->nr, buf->buf.index, buf->hw.vma_nopen);
		v4l2_q_yank_node(&btvc->cap_wait_q, &buf->q);
		v4l2_q_yank_node(&btvc->cap_done_q, &buf->q);

		++buf;
	}
	if(btvc->map<=0)
		capture_unrequest_buffers(btvc);
}

#if  LINUX_VERSION_CODE > KERNEL_VERSION(2,3,0)
struct page *mmap_vma_nopage(struct vm_area_struct *vma, unsigned long address, int write)
#else
static unsigned long mmap_vma_nopage(struct vm_area_struct *vma, unsigned long address, int write)
#endif
{
	struct bttv *btv;
	struct bttvo *btvo;
	struct bttvc *btvc;
	struct stream_buffer *buf;
	unsigned long offset_into_buffer;
#if  LINUX_VERSION_CODE > KERNEL_VERSION(2,3,0)
	struct page *page;
#else
	unsigned long page;
#endif

	int n = 1;
	
	btv = (struct bttv*)v4l2_device_from_file(vma->vm_file);
	btvo = (struct bttvo*)v4l2_openid_from_file(vma->vm_file);
	btvc = &btvo->stream;

	Dprintk(3,"bttv%d: mmap_vma_nopage start\n", btv->nr);
	if ((!btv)||(!btvo))
		return 0;

	if (btvc->cap_contig)
	{
		buf = &btvc->bufs[0];
		n = btvc->req;
	} else
	{
#if  LINUX_VERSION_CODE > KERNEL_VERSION(2,3,0)
		buf = mmap_stream_buffer_from_offset(btvc, vma->vm_pgoff * PAGE_SIZE);
#else
		buf = mmap_stream_buffer_from_offset(btvc, vma->vm_offset);
#endif
	}

	if (buf == NULL)
		return 0;
	offset_into_buffer = address - vma->vm_start;
	if (offset_into_buffer >= btvo->cap_bsize * n)
	{
		Dprintk(1,"bttv%d: Attempt to read past end of mmap() buffer\n", btv->nr);
		return 0;
	}
	page = v4l2_vmalloc_to_page(buf->hw.data + offset_into_buffer);
#if  LINUX_VERSION_CODE > KERNEL_VERSION(2,3,0)
	if (page == NULL)
	{
		printk("bttv%d: mmap_vma_nopage - page == NULL", btv->nr); 
		return 0;
	}
	atomic_inc(&page->count);
#else
	if (page == 0)
		return 0;
	atomic_inc(&mem_map[virt_to_page(__va(page))].count);
#endif
	Dprintk(3,"bttv%d: mmap_vma_nopage success\n", btv->nr);
	return page;
}

static struct vm_operations_struct capture_vma_operations =
{
	open:	mmap_vma_open,
	close:	mmap_vma_close,
	nopage:	mmap_vma_nopage
};

#else

static void mmap_vma_open(struct vm_area_struct *vma)
{
	struct bttv *btv = (struct bttv*)v4l2_device_from_file(vma->vm_file);
	struct bttvo *btvo = (struct bttvo*)v4l2_openid_from_file(vma->vm_file);
	struct bttvc *btvc = &btvo->stream;
 	struct stream_buffer *buf;

	if ((!btv)||(!btvo))
		return;
		
	if (btvc->cap_contig)
		buf = &btvc->bufs[0];
	else
#if  LINUX_VERSION_CODE > KERNEL_VERSION(2,3,0)
		buf = mmap_stream_buffer_from_offset(btvc, vma->vm_pgoff * PAGE_SIZE);
#else
		buf = mmap_stream_buffer_from_offset(btvc, vma->vm_offset);
#endif
	buf->hw.vma_nopen++;
	Dprintk(2,"bttv%d: vma_open called on buffer %d (refcount %d)\n",
		btv->nr, buf->buf.index, buf->hw.vma_nopen);
}


static void mmap_vma_unmap(struct vm_area_struct *vma, unsigned long address, size_t size)
{
	struct bttv *btv = (struct bttv*)v4l2_device_from_file(vma->vm_file);
	struct bttvo *btvo = (struct bttvo*)v4l2_openid_from_file(vma->vm_file);
	struct bttvc *btvc = &btvo->stream;
 	struct stream_buffer *buf;
	int i, n;
	struct bttvc *c;
	
	if((!btv)||(!btvo))
		return;
	if (btvc->cap_contig)
	{
		n = btvc->map;
		buf = &btvc->bufs[0];
	} else
	{
#if  LINUX_VERSION_CODE > KERNEL_VERSION(2,3,0)
		buf = mmap_stream_buffer_from_offset(btvc, vma->vm_pgoff * PAGE_SIZE);
#else
		buf = mmap_stream_buffer_from_offset(btvc, vma->vm_offset);
#endif
		n = 1;
	}
	if(!buf)
	{
		Dprintk(1,"bttv%d: munmap: invalid buffer!\n", btv->nr);
		return;
	}
	buf->hw.vma_nopen--;
	Dprintk(1,"bttv%d: vma_unmap called on buffer %d (refcount %d)\n",
		btv->nr, buf->buf.index, buf->hw.vma_nopen);
	if (buf->hw.vma_nopen > 0)
		return;

	for(c=btv->streaming_caps; c; c=c->next)
		if(c==btvc)
		{
			Dprintk(1,"bttv%d: Warning - final munmap() called while streaming\n", btv->nr);
			capture_streamoff(btvc, V4L2_BUF_TYPE_CAPTURE);
		}

	for (i = 0; i < n; i++)
	{
		v4l2_q_yank_node(&btvc->cap_wait_q, &buf->q);
		v4l2_q_yank_node(&btvc->cap_done_q, &buf->q);

		++buf;
		if (btvc->map > 0)
			--btvc->map;
	}
	if(btvc->map<=0)
		capture_unrequest_buffers(btvc);
}

static struct vm_operations_struct capture_vma_operations =
{
	open:	mmap_vma_open,
	unmap:	mmap_vma_unmap,
};


#endif /* BTTV_MM */

/***********************************/
/* ioctls and supporting functions */
/***********************************/

static int bt848_bright(struct bttv *btv, int *arg)
{
	int bright = *arg;
	btwrite(bright&0xff, BT848_BRIGHT);
	return 0;
}

static int bt848_hue(struct bttv *btv, int *arg)
{
	int hue = *arg;
	btwrite(hue&0xff, BT848_HUE);
	return 0;
}

static int bt848_contrast(struct bttv *btv, int *arg)
{
	unsigned int cont = *arg;
	unsigned int conthi;

	conthi=(cont>>6)&4;
	btwrite(cont&0xff, BT848_CONTRAST_LO);
	btaor(conthi, ~4, BT848_E_CONTROL);
	btaor(conthi, ~4, BT848_O_CONTROL);
	return 0;
}

static void bt848_sat_u(struct bttv *btv, unsigned long data)
{
	u32 datahi;

	datahi=(data>>7)&2;
	btwrite(data&0xff, BT848_SAT_U_LO);
	btaor(datahi, ~2, BT848_E_CONTROL);
	btaor(datahi, ~2, BT848_O_CONTROL);
}

static void bt848_sat_v(struct bttv *btv, unsigned long data)
{
	u32 datahi;

	datahi=(data>>8)&1;
	btwrite(data&0xff, BT848_SAT_V_LO);
	btaor(datahi, ~1, BT848_E_CONTROL);
	btaor(datahi, ~1, BT848_O_CONTROL);
}

static int bt848_sat(struct bttv *btv, int *sat)
{
	bt848_sat_v(btv, *sat);
	bt848_sat_u(btv, *sat);
	return 0;
}

static int bt848_mute(struct bttv *btv, int *arg)
{
	if(*arg)
		audio(btv, AUDIO_MUTE, 1);
	else
		audio(btv, AUDIO_UNMUTE, 1);
	return 0;
}

static int bt848_volume(struct bttv *btv, int *arg)
{
	struct video_audio va;
	bttv_call_i2c_clients(btv, VIDIOCGAUDIO, &va);
	va.volume = *arg;
	bttv_call_i2c_clients(btv, VIDIOCSAUDIO, &va);
	return 0;
}

static int bt848_bass(struct bttv *btv, int *arg)
{
	struct video_audio va;
	bttv_call_i2c_clients(btv, VIDIOCGAUDIO, &va);
	va.bass = *arg;
	bttv_call_i2c_clients(btv, VIDIOCSAUDIO, &va);
	return 0;
}

static int bt848_treble(struct bttv *btv, int *arg)
{
	struct video_audio va;
	bttv_call_i2c_clients(btv, VIDIOCGAUDIO, &va);
	va.treble = *arg;
	bttv_call_i2c_clients(btv, VIDIOCSAUDIO, &va);
	return 0;
}

static int bt848_balance(struct bttv *btv, int *arg)
{
	struct video_audio va;
	bttv_call_i2c_clients(btv, VIDIOCGAUDIO, &va);
	va.balance = *arg;
	bttv_call_i2c_clients(btv, VIDIOCSAUDIO, &va);
	return 0;
}

static int bt848_burst(struct bttv *btv, int *arg)
{
	int i;
	
	if(btv->pv)
		return -EBUSY;
	for(i=0; i<MAX_OPENS; i++)
		if((btv->open_data[i].isopen)&&(btv->open_data[i].read.req)
		   &&(btv->open_data[i].stream.req))
			return -EBUSY;
	for(i=0; i<MAX_VBI_OPENS; i++)
		if((btv->vbi_open_data[i].isopen)&&(btv->vbi_open_data[i].read.req)
		   &&(btv->vbi_open_data[i].stream.req))
			return -EBUSY;
	i = btv->dma;
	if(i)
		bt848_stop_dma(btv);

	tvnorms[0].scaledtwidth = 1135 - (*arg?BURSTOFFSET-2:0);
	tvnorms[0].hdelayx1     = 186  - (*arg?BURSTOFFSET  :0);
	tvnorms[2].scaledtwidth = 1135 - (*arg?BURSTOFFSET-2:0);
	tvnorms[2].hdelayx1     = 186  - (*arg?BURSTOFFSET  :0);
	
	bt848_set_standard(btv);
	if(btv->pvstat==PV_WIN)
		btv->pvstat = PV_FBUF;
	if(btv->pvstat==PV_OVL)
		btv->pvstat = PV_INIT;
	if(i)
		bt848_set_dma(btv);

	return 0;
}

struct control_t
{
	u32 id;
	char name[32];
	int minimum;
	int maximum;
	int step;
	int default_value;
	int value;
	u32 type;
	u32 flags;
	int (*op)(struct bttv *btv, int *arg);
};

#define V4L2_CID_BURST V4L2_CID_PRIVATE_BASE
#define V4L2_CID_PRIVATE_LASTP1 (V4L2_CID_BURST+1)
struct control_t controls[]=
{
 {V4L2_CID_BRIGHTNESS, "Brightness", -128, 127, 1, 0, 0, V4L2_CTRL_TYPE_INTEGER, 0, bt848_bright},
 {V4L2_CID_HUE, "Hue", -128, 127, 1, 0, 0, V4L2_CTRL_TYPE_INTEGER, 0, bt848_hue},
 {V4L2_CID_CONTRAST, "Contrast", 0, 511, 1, 216, 216, V4L2_CTRL_TYPE_INTEGER, 0, bt848_contrast},
 {V4L2_CID_SATURATION, "Saturation", 0, 511, 1, 180, 180, V4L2_CTRL_TYPE_INTEGER, 0, bt848_sat},
 {V4L2_CID_AUDIO_MUTE, "Audio Mute", 0, 1, 1, 1, 1, V4L2_CTRL_TYPE_BOOLEAN, 0, bt848_mute},
 {V4L2_CID_AUDIO_VOLUME, "Audio Volume", 0, 65535, 1, 32768, 32768, V4L2_CTRL_TYPE_INTEGER, V4L2_CTRL_FLAG_DISABLED, bt848_volume},
 {V4L2_CID_AUDIO_BASS, "Audio Bass", 0, 65535, 1, 32768, 32768, V4L2_CTRL_TYPE_INTEGER, V4L2_CTRL_FLAG_DISABLED, bt848_bass},
 {V4L2_CID_AUDIO_TREBLE, "Audio Treble", 0, 65535, 1, 32768, 32768, V4L2_CTRL_TYPE_INTEGER, V4L2_CTRL_FLAG_DISABLED, bt848_treble},
 {V4L2_CID_AUDIO_BALANCE, "Audio Balance", 0, 65535, 1, 32768, 32768, V4L2_CTRL_TYPE_INTEGER, V4L2_CTRL_FLAG_DISABLED, bt848_balance},
 {V4L2_CID_BURST, "Burst", 0, 1, 1, 0, 0, V4L2_CTRL_TYPE_BOOLEAN, 0, bt848_burst},
};

int no_controls=10;

static int find_control(u32 id)
{
	int i;
	for(i=0; i<no_controls; i++)
		if(controls[i].id==id)
			return i;
	return -1;
}

static int do_control(struct bttv *btv, int i, int arg)
{
	if((i<0)||(i>=no_controls))
		return -1;
	if(arg>controls[i].maximum)
		arg = controls[i].maximum;
	if(arg<controls[i].minimum)
		arg = controls[i].minimum;
	if(controls[i].op(btv, &arg)<0)
		return -1;
	controls[i].value = arg;
	return 0;
}

static void default_controls(struct bttv *btv)
{
	int i;
	
	for(i=0; i<no_controls; i++)
		do_control(btv, i, controls[i].default_value);
}

#define  TRITON_PCON	           0x50 
#define  TRITON_BUS_CONCURRENCY   (1<<0)
#define  TRITON_STREAMING	  (1<<1)
#define  TRITON_WRITE_BURST	  (1<<2)
#define  TRITON_PEER_CONCURRENCY  (1<<3)

/* V4L2 device stuff */
/* Serialise everything with btv->lock.  Then the only thing that can */
/* happen out of sequence is interrupts ... */

static int v4l2_open(struct v4l2_device *v, int flags, void **idptr)
{
	struct bttv *btv = (struct bttv*)v->priv;
	int i;
	int ret;
	
	MOD_INC_USE_COUNT;
	down(&btv->lock);

	Dprintk(2,"bttv%d: Open for %s.\n", btv->nr, btv->video_dev.name);
	
	if(btv->radio)
	{
		ret = -EBUSY;
		MOD_DEC_USE_COUNT;
		goto open_done;
	}
	for (i = 0; i < MAX_OPENS; ++i)
		if (!btv->open_data[i].isopen)
			break;
	if (i == MAX_OPENS)/* No available open_data structures */
	{
		Dprintk(2,"bttv%d: No more opens on this device\n", btv->nr);
		ret = -EBUSY;
		MOD_DEC_USE_COUNT;
		goto open_done;
	}
	memset(&btv->open_data[i], 0, sizeof(struct bttvo));
	btv->open_data[i].index = i;
	btv->open_data[i].vbi = 0;
	btv->open_data[i].read.open = &(btv->open_data[i]);
	btv->open_data[i].stream.open = &(btv->open_data[i]);
	btv->open_data[i].isopen = 1;
	btv->open_data[i].input = btv->input;
	btv->open_data[i].dev = v->priv;
	*idptr = &btv->open_data[i];
	if (flags & O_NONCAP) /*  Non-capturing open */
	{
		btv->open_data[i].noncapturing = 1;
	} else
	{
		btv->open_data[i].noncapturing = 0;
		capture_set_format(&btv->open_data[i], &btv->cap_fmt);
		v4l2_q_init(&btv->open_data[i].read.cap_wait_q);
		v4l2_q_init(&btv->open_data[i].read.cap_done_q);
		v4l2_q_init(&btv->open_data[i].stream.cap_wait_q);
		v4l2_q_init(&btv->open_data[i].stream.cap_done_q);
		init_waitqueue_head(&btv->open_data[i].waitq);
	}
	
	bt848_open(btv, 0);

	Dprintk(2,"bttv%d: Open succeeded\n", btv->nr);

	ret = 0;
open_done:	
	up(&btv->lock);
	return ret;
}

static void v4l2_close(void *id)
{
	struct bttvo *btvo = (struct bttvo *)id;
	struct bttv *btv = btvo->dev;

	down(&btv->lock);

	Dprintk(2,"bttv%d: Close for %s.\n", btv->nr, btv->video_dev.name);

	if(!btvo->noncapturing)
	{
		// shut off capture
		capture_streamoff(&btvo->read, V4L2_BUF_TYPE_CAPTURE);
		capture_unrequest_buffers(&btvo->read);
		capture_streamoff(&btvo->stream, V4L2_BUF_TYPE_CAPTURE);
		capture_unrequest_buffers(&btvo->stream);
		Dprintk(2,"bttv%d: capture stopped.\n", btv->nr);
	}
	if(btvo==btv->pv_open)
	{
		preview(btv, 0);
		free_preview(btv);
		btv->pv_open = NULL;
	}

	if(btvo->cap_dummy.risc_even)kfree(btvo->cap_dummy.risc_even);
	if(btvo->cap_dummy.risc_odd)kfree(btvo->cap_dummy.risc_odd);
	if(btvo->cap_dummy.dma)kfree(btvo->cap_dummy.dma);

	btvo->isopen = 0;
	bt848_close(btv, 0);

	up(&btv->lock);
	MOD_DEC_USE_COUNT;
}

static int v4l2_ioctl(void *id, unsigned int cmd, void *arg)
{
	struct bttvo *btvo = (struct bttvo *)id;
	struct bttv *btv = (struct bttv *)btvo->dev;
	struct bttvc *c;
	int ret = -ENOIOCTLCMD;
	
	down(&btv->lock);
	
	Dprintk(2,"bttv%d: ioctl() for %s \n", btv->nr, btv->video_dev.name);
	
	switch (cmd)
	{
		case VIDIOC_QUERYCAP:
			{
				struct v4l2_capability *cap = arg;
				sprintf(cap->name, "Bttv2(%d) - Video", btv->nr);
				cap->type = V4L2_TYPE_CAPTURE;
				cap->inputs = bttv_tvcards[btv->type].video_inputs;
				cap->outputs = 0;
				cap->audios = bttv_tvcards[btv->type].audio_inputs;
				cap->maxwidth = tvnorms[btv->norm].maxwidth;
				cap->maxheight = tvnorms[btv->norm].sheight;
				cap->minwidth = 16;
				cap->minheight = 16;
				cap->maxframerate = 60;
				cap->flags = V4L2_FLAG_PREVIEW|V4L2_FLAG_TUNER|V4L2_FLAG_STREAMING|V4L2_FLAG_SELECT|V4L2_FLAG_READ;
				ret = 0;
				goto ioctl_done;
			}
		case VIDIOC_ENUMSTD:
			{
				struct v4l2_enumstd *e = arg;
				
				if((e->index<0)||(e->index>=TVNORMS))
				{
					ret = -EINVAL;
					goto ioctl_done;
				}
				v4l2_video_std_construct(&e->std, tvnorms[e->index].v4l2_id, 0);
				e->inputs = 0xffffffffUL;
				e->outputs = 0;
				ret = 0;
				goto ioctl_done;
			}
		case VIDIOC_G_STD:
			{
				struct v4l2_standard *s = arg;
				v4l2_video_std_construct(s, tvnorms[btv->norm].v4l2_id, 0);
				ret = 0;
				goto ioctl_done;
			}
		case VIDIOC_S_STD:
			{
				struct v4l2_standard *s = arg;
				int i;
				int id = v4l2_video_std_confirm(s);
				
				Dprintk(2,"bttv%d: S_STD called!\n", btv->nr);
				
				if(btv->pv)
				{
					ret = -EBUSY;
					goto ioctl_done;
				}
				for(i=0; i<MAX_OPENS; i++)
					if((btv->open_data[i].isopen)&&
					    ((btv->open_data[i].stream.req)||(btv->open_data[i].read.req)))
					{
						ret = -EBUSY;
						goto ioctl_done;
					}  
				for(i=0; i<MAX_VBI_OPENS; i++)
					if((btv->vbi_open_data[i].isopen)&&
					    ((btv->vbi_open_data[i].stream.req)||(btv->vbi_open_data[i].read.req)))
					{
						ret = -EBUSY;
						goto ioctl_done;
					}  
				
				for(i=0; i<TVNORMS; i++)
				{
					if(tvnorms[i].v4l2_id==id)
					{
						bt848_stop_dma(btv);
						btv->norm = i;
						bt848_set_standard(btv);
						// Invalidate any geometries
						if(btv->pvstat==PV_WIN)
							btv->pvstat = PV_FBUF;
						if(btv->pvstat==PV_OVL)
							btv->pvstat = PV_INIT;
						bt848_set_dma(btv);
						ret = 0;
						goto ioctl_done;
					}
				}
				
				ret = -EINVAL;
				goto ioctl_done;
			}
		/* I am not sure about these - duplicated in TUNER, but still needed for v4l1 compat */
		case VIDIOC_S_AUDIO:
		case VIDIOC_G_AUDIO:
			{
				struct v4l2_audio *a = arg;
				if((a->audio<0)||(a->audio>=bttv_tvcards[btv->type].audio_inputs))
				{
					ret = -EINVAL;
					goto ioctl_done;
				}
				sprintf(a->name, "Audio %d", a->audio);
				a->capability = 0;
				a->mode = 0;
				ret = 0;
				goto ioctl_done;
			}
		case VIDIOC_G_TUNER:
			{
				struct v4l2_tuner *t = arg;
				struct video_audio va;
				
				t->input = bttv_tvcards[btv->type].tuner;
				sprintf(t->name, "Tuner");
				v4l2_video_std_construct(&t->std, tvnorms[btv->norm].v4l2_id, 0);
				bttv_call_i2c_clients(btv, VIDIOCGAUDIO, &va);
				/* card specific hooks */
				if (bttv_tvcards[btv->type].audio_hook)
					bttv_tvcards[btv->type].audio_hook(btv,&va,0);
				t->capability = V4L2_TUNER_CAP_NORM;
				if(va.flags&VIDEO_SOUND_STEREO)
					t->capability|=V4L2_TUNER_CAP_STEREO;
				if(va.flags&VIDEO_SOUND_LANG1)
					t->capability|=V4L2_TUNER_CAP_LANG1;
				if(va.flags&VIDEO_SOUND_LANG2)
					t->capability|=V4L2_TUNER_CAP_LANG2;
				t->rangelow = 0;
				t->rangehigh = 0xffffffffUL;
				t->rxsubchans = 0xffffffffUL;
				t->audmode = 0;
				if(va.mode&VIDEO_SOUND_MONO)
					t->audmode|=V4L2_TUNER_MODE_MONO;
				if(va.mode&VIDEO_SOUND_STEREO)
					t->audmode|=V4L2_TUNER_MODE_STEREO;
				if(va.mode&VIDEO_SOUND_LANG1)
					t->audmode|=V4L2_TUNER_MODE_LANG1;
				if(va.mode&VIDEO_SOUND_LANG2)
					t->audmode|=V4L2_TUNER_MODE_LANG2;
				t->signal = (btread(BT848_DSTATUS)&BT848_DSTATUS_HLOC) ? 0xFFFF : 0;
				t->afc = 0;
				ret = 0;
				goto ioctl_done;
			}
		case VIDIOC_S_TUNER:
			{
				struct v4l2_tuner *t = arg;
				struct video_audio va;
				
				if(t->input!=bttv_tvcards[btv->type].tuner)
				{
					ret = -EINVAL;
					goto ioctl_done;
				}
				bttv_call_i2c_clients(btv, VIDIOCGAUDIO, &va);
				/* card specific hooks */
				if (bttv_tvcards[btv->type].audio_hook)
					bttv_tvcards[btv->type].audio_hook(btv,&va,0);
				va.mode = 0;
				if(t->audmode&V4L2_TUNER_MODE_MONO)
					va.mode|=VIDEO_SOUND_MONO;
				if(t->audmode&V4L2_TUNER_MODE_STEREO)
					va.mode|=VIDEO_SOUND_STEREO;
				if(t->audmode&V4L2_TUNER_MODE_LANG1)
					va.mode|=VIDEO_SOUND_LANG1;
				if(t->audmode&V4L2_TUNER_MODE_LANG2)
					va.mode|=VIDEO_SOUND_LANG2;
				bttv_call_i2c_clients(btv, VIDIOCSAUDIO, &va);
				/* card specific hooks */
				if (bttv_tvcards[btv->type].audio_hook)
					bttv_tvcards[btv->type].audio_hook(btv,&va,0);
				ret = 0;
				goto ioctl_done;
			}
		case VIDIOC_ENUMINPUT:
			{
				struct v4l2_input *i = arg;
				
				if(i->index>=bttv_tvcards[btv->type].video_inputs)
				{
					ret = -EINVAL;
					goto ioctl_done;
				}
				if(i->index==bttv_tvcards[btv->type].tuner)
				{
					sprintf(i->name, "Tuner");
					i->type = V4L2_INPUT_TYPE_TUNER;
					i->capability = V4L2_INPUT_CAP_AUDIO;
					i->assoc_audio = 0;
					ret = 0;
					goto ioctl_done;
				}
				if(i->index==bttv_tvcards[btv->type].svhs)
				{
					sprintf(i->name, "Svideo");
				} else
				{
					int j = i->index;
					if(i->index>bttv_tvcards[btv->type].tuner)
						j--;
					if(i->index>bttv_tvcards[btv->type].svhs)
						j--;
					sprintf(i->name, "Composite %d", j);
				}
				i->type = V4L2_INPUT_TYPE_CAMERA;
				i->capability = 0;
				i->assoc_audio = -1;
			
				ret = 0;
				goto ioctl_done;
			}
		case VIDIOC_S_INPUT:
			{
				int *i = arg;
				
				if((*i<0)||(*i>=bttv_tvcards[btv->type].video_inputs))
				{
					ret = -EINVAL;
					goto ioctl_done;
				}
				btvo->input = *i;
				if( (!btv->streaming_caps)&&
				    (!btv->streaming_vbis) )
					bt848_set_input(btv, *i);
				ret = 0;
				goto ioctl_done;
			}
		case VIDIOC_G_INPUT:
			{
				int *i = arg;
				*i = btv->input;
				ret = 0;
				goto ioctl_done;
			}
		case VIDIOC_S_FREQ:
			{
				u32 *f = arg;
				btv->freq = *f;
				bttv_call_i2c_clients(btv,VIDIOCSFREQ,f);
				ret = 0;
				goto ioctl_done;
			}
		case VIDIOC_G_FREQ:
			{
				u32 *f = arg;
				*f = btv->freq;
				ret = 0;
				goto ioctl_done;
			}
		case VIDIOC_QUERYMENU:
			{
				ret = -EINVAL; // no menus here!
				goto ioctl_done;
			}
		case VIDIOC_QUERYCTRL:
			{
				struct v4l2_queryctrl *c = arg;
				int i;
				struct video_audio va;
				
				va.flags = 0;
				bttv_call_i2c_clients(btv, VIDIOCGAUDIO, &va);
				/* card specific hooks */
				if (bttv_tvcards[btv->type].audio_hook)
					bttv_tvcards[btv->type].audio_hook(btv,&va,0);
				if(va.flags&VIDEO_AUDIO_VOLUME)
					controls[find_control(V4L2_CID_AUDIO_VOLUME)].flags = 0;
				if(va.flags&VIDEO_AUDIO_BASS)
					controls[find_control(V4L2_CID_AUDIO_BASS)].flags = 0;
				if(va.flags&VIDEO_AUDIO_TREBLE)
					controls[find_control(V4L2_CID_AUDIO_TREBLE)].flags = 0;
				// Can't test for this, so leave it on 
				controls[find_control(V4L2_CID_AUDIO_BALANCE)].flags = 0;
				
				if(((c->id>=V4L2_CID_BASE)&&(c->id<V4L2_CID_LASTP1))
				   ||((c->id>=V4L2_CID_PRIVATE_BASE)&&(c->id<V4L2_CID_PRIVATE_LASTP1)))
				{
					if((i=find_control(c->id))<=0)
					{
						memset(c, 0, sizeof(struct v4l2_queryctrl));
						sprintf(c->name, "None");
						c->flags = V4L2_CTRL_FLAG_DISABLED;
					} else
					{
						sprintf(c->name, "%s", controls[i].name);
						c->minimum = controls[i].minimum;
						c->maximum = controls[i].maximum;
						c->step = controls[i].step;
						c->default_value = controls[i].default_value;
						c->type = controls[i].type;
						c->flags = controls[i].flags;
					}
					ret = 0;
					goto ioctl_done;
				}
				ret = -EINVAL;
				goto ioctl_done;
			}
		case VIDIOC_G_CTRL:
			{
				struct v4l2_control *c = arg;
				int i;
				
				if(((c->id>=V4L2_CID_BASE)&&(c->id<V4L2_CID_LASTP1))
				   ||((c->id>=V4L2_CID_PRIVATE_BASE)&&(c->id<V4L2_CID_PRIVATE_LASTP1)))
				{
					if((i=find_control(c->id))<=0)
					{
						ret = -EINVAL;
						goto ioctl_done;
					} else
					{
						c->value = controls[i].value;
					}
					ret = 0;
					goto ioctl_done;
				}
				ret = -EINVAL;
				goto ioctl_done;
			}
		case VIDIOC_S_CTRL:
			{
				struct v4l2_control *c = arg;
				int i;
				
				if(((c->id>=V4L2_CID_BASE)&&(c->id<V4L2_CID_LASTP1))
				   ||((c->id>=V4L2_CID_PRIVATE_BASE)&&(c->id<V4L2_CID_PRIVATE_LASTP1)))
				{
					if((i=find_control(c->id))<=0)
					{
						ret = -EINVAL;
						goto ioctl_done;
					} else
					{
						if(do_control(btv, i, c->value)<0)
						{
							ret = -EBUSY;
							goto ioctl_done;
						}
					}
					ret = 0;
					goto ioctl_done;
				}
				ret = -EINVAL;
				goto ioctl_done;
			}
		case VIDIOC_ENUM_PIXFMT:
		case VIDIOC_ENUM_FBUFFMT:
			{
				struct v4l2_fmtdesc *f = arg;
				
				if((f->index>=format_max)||(f->index<0))
				{
					ret = -EINVAL;
					goto ioctl_done;
				}
				sprintf(f->description, "%s", formats[f->index].description);
				f->pixelformat = formats[f->index].pixelformat;
				f->flags = formats[f->index].flags;
				f->depth = formats[f->index].depth;
				ret = 0;
				goto ioctl_done;
			}
		case VIDIOC_S_FBUF:
			{
				struct v4l2_framebuffer *fb = arg;
				int i;
				
				if(!capable(CAP_SYS_ADMIN))
				{
					Dprintk(1,"bttv%d: Only root can set framebuffer!\n", btv->nr);
					ret = -EPERM;
					goto ioctl_done;
				}
				if(!btv->pv_open)
					btv->pv_open = btvo;
				if(btv->pv_open!=btvo)
				{
					Dprintk(1,"bttv%d: s_fbuf - pv is busy\n", btv->nr);
					ret = -EBUSY;
					goto ioctl_done;
				}
				i = find_format(fb->fmt.pixelformat);
				if(i<0)
				{
					printk("bttv%d: framebuffer pixelformat not supported!\n", btv->nr);
					ret = -EINVAL;
					goto ioctl_done;
				}
				memcpy(&btv->fb, fb, sizeof(struct v4l2_framebuffer));
				btv->fb.fmt.depth = formats[i].depth;
				Dprintk(2,"bttv%d: s_fbuf base: %u %u\n", btv->nr, (u32)btv->fb.base[0], (u32)fb->base[0]);
				if(!(btv->fb.fmt.flags&V4L2_FMT_FLAG_BYTESPERLINE))
				{
					Dprintk(1,"bttv%d: Setting bytesperline from width.\n", btv->nr);
					btv->fb.fmt.bytesperline = (btv->fb.fmt.width * formats[i].depth + 7)>>3;
				}
				if(btv->fb.flags&V4L2_FBUF_FLAG_PRIMARY)
				{
					btv->pvstat = PV_FBUF;
					ret = 0;
					goto ioctl_done;
				}
				if(btv->fb.flags&V4L2_FBUF_FLAG_OVERLAY)
				{
					btv->win.x = 0;
					btv->win.y = 0;
					btv->win.width = MIN(tvnorms[btv->norm].maxwidth, btv->fb.fmt.width);
					btv->win.height = MIN(tvnorms[btv->norm].sheight, btv->fb.fmt.height);
					btv->win.clips = NULL;
					btv->win.clipcount = -1;
					if((ret=pv_new_win(btv, NULL))<0)
					{
						Dprintk(1,"bttv%d: could not set new window!\n", btv->nr);
						goto ioctl_done;
					}
					btv->pvstat = PV_OVL;
					ret = 0;
					goto ioctl_done;
				}
				Dprintk(1,"bttv%d: no framebuffer flags set, assuming primary!\n", btv->nr);
				btv->pvstat = PV_FBUF;
				ret = 0;
				goto ioctl_done;  
			}
		case VIDIOC_G_FBUF:
			{
				struct v4l2_framebuffer *fb = arg;
				memcpy(fb, &btv->fb, sizeof(struct v4l2_framebuffer));
				Dprintk(2,"bttv%d: g_fbuf base: %u %u\n", btv->nr, (u32)btv->fb.base[0], (u32)fb->base[0]);
				fb->capability = V4L2_FBUF_CAP_CLIPPING | V4L2_FBUF_CAP_BITMAP_CLIPPING;
				ret = 0;
				goto ioctl_done;
			}
		case VIDIOC_S_WIN:
			{
				struct v4l2_window *win = arg;
				int i;
				u8 *save_bitmap;

				if((btv->pvstat!=PV_WIN)&&(btv->pvstat!=PV_FBUF))
				{
					Dprintk(1,"bttv%d: fb not set, or in overlay mode!\n", btv->nr);
					ret = -EINVAL;
					goto ioctl_done;
				}
				if(!btv->pv_open)
					btv->pv_open = btvo;
				if(btv->pv_open!=btvo)
				{
					Dprintk(1,"bttv%d: s_win - pv is busy\n", btv->nr);
					ret = -EBUSY;
					goto ioctl_done;
				}
				if(btv->pv)
				{
					i = 1;
					preview(btv, 0);
				} else i=0;
				
				save_bitmap = btv->win.bitmap;
				memcpy(&btv->win, win, sizeof(struct v4l2_window));
				btv->win.bitmap = save_bitmap;
				if((ret=pv_new_win(btv, win->bitmap))<0)
				{
					Dprintk(1,"bttv%d: could not set new window!\n", btv->nr);
					btv->pvstat = PV_FBUF;
					goto ioctl_done;
				}
				btv->pvstat = PV_WIN;
				if(i)
					preview(btv, 1);
				ret = 0;
				goto ioctl_done;
			}
		case VIDIOC_G_WIN:
			{
				memcpy(arg, &btv->win, sizeof(struct v4l2_window));
				ret = 0;
				goto ioctl_done;
			}
		case VIDIOC_PREVIEW:
			{
				__u32 *i = arg;
				
				if(!btv->pv_open)
					btv->pv_open = btvo;
				if(btv->pv_open!=btvo)
				{
					Dprintk(1,"bttv%d: preview - pv is busy\n", btv->nr);
					ret = -EBUSY;
					goto ioctl_done;
				}
				ret = preview(btv, *i);
				goto ioctl_done;
			}
		case VIDIOC_S_FMT:
			{
				struct v4l2_format *f = arg;
				if (btvo->noncapturing)
				{
					Dprintk(1,"bttv%d: S_FMT illegal on noncap open.\n", btv->nr);
					ret = -EPERM;
					goto ioctl_done;
				}
				for(c=btv->streaming_caps; c; c=c->next)
					if(c->open==btvo)
					{
						Dprintk(1,"bttv%d: S_FMT while streaming.\n", btv->nr);
						ret = -EBUSY;
						goto ioctl_done;
					}
/*				if(f->type!=V4L2_BUF_TYPE_CAPTURE)
				{
					ret = -EINVAL;
					goto ioctl_done;
				}*/
				ret = capture_set_format(btvo, &f->fmt.pix);
				goto ioctl_done;
			}
		case VIDIOC_G_FMT:
			{
				struct v4l2_format *f = arg;
				if(f->type!=V4L2_BUF_TYPE_CAPTURE)
				{
					ret = -EINVAL;
					goto ioctl_done;
				}
				memcpy(&f->fmt.pix, &btvo->cap_fmt, sizeof(struct v4l2_pix_format));
				ret = 0;
				goto ioctl_done;
			}
		case VIDIOC_REQBUFS:
			{
				struct v4l2_requestbuffers *req = arg;
				if (btvo->noncapturing)
				{
					Dprintk(1,"bttv%d: S_FMT illegal on noncap open.\n", btv->nr);
					ret = -EPERM;
					goto ioctl_done;
				}
				ret = capture_request_buffers(&btvo->stream, req);
				goto ioctl_done;
			}
		case VIDIOC_QUERYBUF:
			{
				struct v4l2_buffer *b = arg;
				
				if (btvo->noncapturing)
				{
					Dprintk(1,"bttv%d: S_FMT illegal on noncap open.\n", btv->nr);
					ret = -EPERM;
					goto ioctl_done;
				}
				if((b->index<0)||(b->index>=btvo->stream.req))
				{
					ret = -EINVAL;
					goto ioctl_done;
				}
				if((b->type & V4L2_BUF_TYPE_field)!=(btvo->stream.bufs[b->index].buf.type & V4L2_BUF_TYPE_field))
				{
					ret = -EINVAL;
					goto ioctl_done;
				}
				memcpy(b, &btvo->stream.bufs[b->index].buf, sizeof(struct v4l2_buffer));
				Dprintk(2,"bttv%d: querybuf %d: %d %d\n", btv->nr, b->index, b->length, b->offset);
				ret = 0;
				goto ioctl_done;
			}
		case VIDIOC_QBUF:
			{
				struct v4l2_buffer *buf = arg;
				if (btvo->noncapturing)
				{
					Dprintk(1,"bttv%d: QBUF illegal in non-capturing open\n", btv->nr);
					ret = -EPERM;
					goto ioctl_done;
				}
				if (!btvo->stream.map)
				{
					Dprintk(1,"bttv%d: QBUF no buffers are mapped\n", btv->nr);
					ret = -EINVAL;
					goto ioctl_done;
				}
				ret = capture_queuebuffer(&btvo->stream, buf);
				goto ioctl_done;
			}
		case VIDIOC_DQBUF:
			{
				struct v4l2_buffer *buf = arg;
				if (btvo->noncapturing)
				{
					Dprintk(1,"bttv%d: DQBUF illegal in non-capturing open\n", btv->nr);
					ret = -EPERM;
					goto ioctl_done;
				}
				ret = capture_dequeuebuffer(&btvo->stream, buf);
				goto ioctl_done;
			}
		case VIDIOC_STREAMON:
			{
				__u32 *type = arg;
				if (btvo->noncapturing)
				{
					Dprintk(1,"btv%d: STREAMON illegal in non-capturing open\n", btv->nr);
					ret = -EPERM;
					goto ioctl_done;
				}
				if (btvo->stream.map!=btvo->stream.req)
				{
					Dprintk(1,"btv%d: no all buffers mapped!\n", btv->nr);
					ret = -EBUSY;
					goto ioctl_done;
				}
				ret = capture_streamon(&btvo->stream, *type);
				goto ioctl_done;
			}
		case VIDIOC_STREAMOFF:
			{
				__u32	*type = arg;
				if (btvo->noncapturing)
				{
					Dprintk(1,"bttv%d: STREAMOFF illegal in non-capturing open\n", btv->nr);
					ret = -EPERM;
					goto ioctl_done;
				}
				capture_streamoff(&btvo->stream, *type);
				ret = 0;
				goto ioctl_done;
			}
		case VIDIOC_G_PERF:
			{
				memcpy(arg, &btvo->perf, sizeof(struct v4l2_performance));
				ret = 0;
				goto ioctl_done;
			}
		case VIDIOC_G_PARM:
		case VIDIOC_S_PARM:
			{
				struct v4l2_streamparm *p = arg;
				p->type = V4L2_BUF_TYPE_CAPTURE;
				p->parm.capture.capability = 0;
				p->parm.capture.capturemode = 0;
				p->parm.capture.timeperframe = btv->tpf;
				ret = 0;
				goto ioctl_done;
			}
	}
ioctl_done:
	up(&btv->lock);
	
	return ret;
}                      

static long v4l2_write(void *id, const char *buf, unsigned long count, int noblock)
{
	struct bttvo *btvo = (struct bttvo *)id;
	struct bttv *btv = (struct bttv *)btvo->dev;

	down(&btv->lock);
	
	Dprintk(1,"bttv%d: Write() not handled for %s.\n", btv->nr, btv->video_dev.name);
	
	up(&btv->lock);
	
	return -EINVAL;
}

static int v4l2_mmap(void *id, struct vm_area_struct *vma)
{
	struct bttvo *btvo = (struct bttvo *)id;
	struct bttv *btv = btvo->dev;
	struct bttvc *btvc = &btvo->stream;
	struct stream_buffer *buf;
	int i, n = 1, ret;
	unsigned long start=(unsigned long)vma->vm_start;

	down(&btv->lock);
	
	Dprintk(2,"bttv%d: mmap() for %s.\n", btv->nr, btv->video_dev.name);

	if (btvo->noncapturing)
	{
		Dprintk(1,"bttv%d: mmap() called on non-capturing open\n", btv->nr);
		ret = -ENODEV;
		goto mmap_done;
	}

#if  LINUX_VERSION_CODE > KERNEL_VERSION(2,3,0)
	buf = mmap_stream_buffer_from_offset(btvc, vma->vm_pgoff * PAGE_SIZE);
#else
	buf = mmap_stream_buffer_from_offset(btvc, vma->vm_offset);
#endif
	if (btvc->cap_contig)
	{
		buf = &btvc->bufs[0];
		n = btvc->req;
	}
	if (buf == NULL)
	{
		Dprintk(1,"bttv%d: mmap() Invalid offset parameter\n", btv->nr);
		ret = -EINVAL;/* no such buffer */
		goto mmap_done;
	}
	if ((unsigned long)btvo->cap_bsize * (unsigned long)n != (unsigned long)(vma->vm_end-vma->vm_start))
	{
		Dprintk(1,"bttv%d: mmap() Wrong length parameter %u %lu\n", btv->nr, btvo->cap_bsize, vma->vm_end - vma->vm_start);
		ret = -EINVAL;/* wrong length */
		goto mmap_done;
	}
	if(!btvc->bufs[0].hw.data) /* If we have not yet malloc'ed data, do it now */
	{
		Dprintk(2,"bttv%d: vmalloc() buffers\n", btv->nr); 
		btvc->bufs[0].hw.data = rvmalloc(btvo->cap_bsize * btvc->req);
		if(!btvc->bufs[0].hw.data)
		{
			printk("bttv%d: mmap() failed to malloc() buffer\n", btv->nr);
			ret = -ENOMEM;
			goto mmap_done;
		}
		btvc->cap_vmalloc = btvo->cap_bsize * btvc->req;
	}
	for (i = 0; i < n; ++i)
	{
#ifdef BTTV_MM
		unsigned long page,pos;
		unsigned long size; 
#endif /* BTTV_MM */

		if (buf->buf.flags & V4L2_BUF_FLAG_MAPPED)
		{
			Dprintk(1,"bttv%d: mmap() Buffer is already mapped\n", btv->nr);
			ret = -EINVAL;/* already mapped */
			goto mmap_done;
		}
		if (buf != &btvc->bufs[0])
			buf->hw.data = buf[-1].hw.data + btvo->cap_bsize;
#ifdef BTTV_MM
		pos=(unsigned long) buf->hw.data;
		size = btvo->cap_bsize;
		while (size > 0)
		{
			page = kvirt_to_pa(pos);
			if (remap_page_range(start, page, PAGE_SIZE, PAGE_SHARED))
				return -EAGAIN;
			start+=PAGE_SIZE;
			pos+=PAGE_SIZE; 
			size-=PAGE_SIZE;
		}
#endif /* BTTV_MM */
		buf->buf.flags |= V4L2_BUF_FLAG_MAPPED;
		++btvc->map;
		++buf;
	}
// #ifndef BTTV_MM
	vma->vm_ops = &capture_vma_operations;
	if (vma->vm_ops->open)
		vma->vm_ops->open(vma);
// #endif
	ret = 0;
mmap_done:	
	Dprintk(2,"bttv%d: mmap() %s: %d\n", btv->nr, btv->video_dev.name, ret);
	up(&btv->lock);
	return ret;
}

static int v4l2_poll(void *id, struct file *file, poll_table *table)
{
	struct bttvo *btvo = (struct bttvo *)id;
	struct bttv *btv = (struct bttv *)btvo->dev;
	struct bttvc *c;
	void *node;
	unsigned long flags;
	
	if (btvo->noncapturing)
	{
		Dprintk(1,"bttv%d: poll() illegal in non-capturing open\n", btv->nr);
		return POLLERR;
	}
	if(btvo->vbi)
	{
		for(c=btv->streaming_vbis; c; c=c->next)
			if(c->open==btvo)
				break;
	} else
	{
		for(c=btv->streaming_caps; c; c=c->next)
			if(c->open==btvo)
				break;
	}
	if(!c)
		return POLLERR;
	
	poll_wait(file, &btvo->waitq, table);
	spin_lock_irqsave(&btv->s_lock, flags);
	node = v4l2_q_peek_head(&btvo->stream.cap_done_q);
	if(!node)
		node = v4l2_q_peek_head(&btvo->read.cap_done_q);
	if (node != NULL)
	{
		spin_unlock_irqrestore(&btv->s_lock, flags);
		return (POLLIN | POLLRDNORM);/* data is ready now */
	}
	node = v4l2_q_peek_head(&btvo->stream.cap_wait_q);
	if(!node)
		node = v4l2_q_peek_head(&btvo->read.cap_done_q);
	if (node == NULL &&
		btvo->stream.wait_even == NULL &&
		btvo->stream.wait_odd == NULL &&
		btvo->read.wait_even == NULL &&
		btvo->read.wait_odd == NULL)
	{
		Dprintk(1,"bttv%d: poll() with no queued buffers\n", btv->nr);
		spin_unlock_irqrestore(&btv->s_lock, flags);
		return POLLERR; /* nothing Qed */
	}
	spin_unlock_irqrestore(&btv->s_lock, flags);
	return 0;
}

static long v4l2_read(void *id, char *buf, unsigned long len, int noblock)
{
	struct bttvo *btvo = (struct bttvo *)id;
	struct bttv *btv = btvo->dev;
	struct bttvc *btvc = &btvo->read;
	long my_timeout;
	int ret;
	int i;
	DECLARE_WAITQUEUE(wait, current);
	
	down(&btv->lock);
	
	Dprintk(2,"bttv%d: read() for %s.\n", btv->nr, btv->video_dev.name);

	if(btvo->noncapturing)
	{
		Dprintk(1,"bttv%d: read(): illegal on noncap open.\n", btv->nr);
		ret = -EINVAL;
		goto read_done;
	}
	if(!btvo->reading)
	{
		struct v4l2_requestbuffers req;
		
		if(btvc->req)
		{
			Dprintk(1,"bttv%d: read(): capture buffers already requested.\n", btv->nr);
			ret = -EBUSY;
			goto read_done;
		}
		req.type = V4L2_BUF_TYPE_CAPTURE;
		req.count = 1;
		if(btvo->vbi)
			ret = vbi_request_buffers(btvc, &req);
		else
			ret = capture_request_buffers(btvc, &req);
		if(ret!=0)
		{
			printk("bttv%d: read(): could not request buffers.\n", btv->nr);
			goto read_done;
		}
		if(btvo->vbi)
			btvc->bufs[0].hw.data = kmalloc(btvo->vbi_bsize, GFP_KERNEL);
		else
			btvc->bufs[0].hw.data = rvmalloc(btvo->cap_bsize);
		if(!btvc->bufs[0].hw.data)
		{
			printk("bttv%d: read(): could not allocate capture memory.\n", btv->nr);
			if(btvo->vbi)
				vbi_unrequest_buffers(btvc);
			else
				capture_unrequest_buffers(btvc);
			ret = -ENOMEM;
			goto read_done;
		}
		if(!btvo->vbi)
			btvc->cap_vmalloc = btvo->cap_bsize;
		btvo->read_buffer.type = V4L2_BUF_TYPE_CAPTURE;
		btvo->read_buffer.index = 0;
		btvc->bufs[0].buf.flags |= V4L2_BUF_FLAG_MAPPED;
		if(btvo->vbi)
		{
			vbi_queuebuffer(btvc, &btvo->read_buffer);
			vbi_streamon(btvc, V4L2_BUF_TYPE_CAPTURE);
		}
		else
		{
			capture_queuebuffer(btvc, &btvo->read_buffer);
			capture_streamon(btvc, V4L2_BUF_TYPE_CAPTURE);
		}
		btvo->reading = 1;
	}
	if(btvo->reading) // read for this device
	{
		btvo->read_buffer.type = V4L2_BUF_TYPE_CAPTURE;
		if(vbi)
			i=vbi_dequeuebuffer(btvc, &btvo->read_buffer);
		else
			i=capture_dequeuebuffer(btvc, &btvo->read_buffer);
		if((i<0)&&(noblock))
		{
			ret = -EWOULDBLOCK;
			goto read_done;
		}
		my_timeout = HZ;
		add_wait_queue(&btvo->waitq, &wait);
		current->state = TASK_INTERRUPTIBLE;
		do
		{
			up(&btv->lock);
			my_timeout = schedule_timeout(my_timeout);
			if(signal_pending(current))
			{
				remove_wait_queue(&btvo->waitq, &wait);
				current->state = TASK_RUNNING;
				ret = -EINTR;
				goto read_done;
			}
			if(my_timeout<=0)
			{
				remove_wait_queue(&btvo->waitq, &wait);
				current->state = TASK_RUNNING;
				Dprintk(1,"bttv%d: Timeout on read().\n", btv->nr);
				ret = 0;
				goto read_done;
			}
			down(&btv->lock);
			if(vbi)
				i=vbi_dequeuebuffer(btvc, &btvo->read_buffer);
			else
				i=capture_dequeuebuffer(btvc, &btvo->read_buffer);
		} while(i<0);
		remove_wait_queue(&btvo->waitq, &wait);
		current->state = TASK_RUNNING;
		if(len < btvo->read_buffer.bytesused)
		{
			ret = -EINVAL;
			goto read_done;
		}
		copy_to_user(buf, btvc->bufs[0].hw.data, btvo->read_buffer.bytesused);
		ret = btvo->read_buffer.bytesused;
		if(vbi)
		{
			vbi_streamoff(btvc, V4L2_BUF_TYPE_CAPTURE);
			vbi_unrequest_buffers(btvc);
		} else
		{
			capture_streamoff(btvc, V4L2_BUF_TYPE_CAPTURE);
			capture_unrequest_buffers(btvc);
		} 
		btvo->reading = 0;
		goto read_done;
	}
	ret = -EBUSY;
read_done:
	up(&btv->lock);
	return ret;
}

static int v4l2_init_done(struct v4l2_device *v)
{
	return 0;
}

static int new_video_device(struct bttv *btv)
{
	struct v4l2_device *v = &btv->video_dev;
	// name is entered by idcard()
	// sprintf(v->name, "Bttv2 captrure device %d", btv->nr);
	v->type = V4L2_TYPE_CAPTURE;
	v->minor = bttv_video[btv->nr];
	v->open = v4l2_open;
	v->close = v4l2_close;
	v->read = v4l2_read;
	v->write = v4l2_write;
	v->ioctl = v4l2_ioctl;
	v->mmap = v4l2_mmap;
	v->poll = v4l2_poll;
	v->initialize = v4l2_init_done;
	v->priv = btv;
	return v4l2_register_device(v);
}; 

/********************************/
/*      RADIO                   */
/********************************/

static int v4l2_radio_open(struct v4l2_device *v, int flags, void **idptr)
{
	struct bttv *btv = (struct bttv*)v->priv;
	unsigned long f;
	int ret;

	MOD_INC_USE_COUNT;
	Dprintk(2,"bttv%d: Radio open for %s.\n", btv->nr, btv->radio_dev.name);
	down(&btv->lock);

	if((btv->open_count)||(btv->radio))
	{
		ret = -EBUSY;
		MOD_DEC_USE_COUNT;
		goto radio_open_done;
	}
	btv->radio = 1;
	f = 400*16;
	bttv_call_i2c_clients(btv,VIDIOCSFREQ,&v);
	bttv_call_i2c_clients(btv,AUDC_SET_RADIO,&btv->tuner_type);
	bt848_muxsel(btv,0);

	btv->open_data[0].noncapturing = 1;
	btv->open_data[0].isopen = 1;
	btv->open_data[0].dev = v->priv;
	*idptr = &btv->open_data[0];
	
	ret = 0;
radio_open_done:
	up(&btv->lock);
	return ret;
}

static void v4l2_radio_close(void *id)
{
	struct bttvo *btvo = (struct bttvo *)id;
	struct bttv *btv = (struct bttv*)btvo->dev;
	
	down(&btv->lock);

	Dprintk(2,"bttv%d: Radio close for %s.\n", btv->nr, btv->radio_dev.name);

	btv->radio = 0;
	up(&btv->lock);
	MOD_DEC_USE_COUNT;
}

static int v4l2_radio_ioctl(void *id, unsigned int cmd, void *arg)
{
	struct bttvo *btvo = (struct bttvo *)id;
	struct bttv *btv = (struct bttv *)btvo->dev;
	int ret;
	
	down(&btv->lock);
	switch(cmd)
	{
		case VIDIOC_QUERYCAP:
			{
				struct v4l2_capability *cap = arg;
				sprintf(cap->name, "Bttv2(%d) - Radio", btv->nr);
				cap->type = V4L2_TYPE_RADIO;
				cap->inputs = 0;
				cap->outputs = 0;
				cap->audios = 1;
				cap->maxwidth = 0;
				cap->maxheight = 0;
				cap->minwidth = 0;
				cap->minheight = 0;
				cap->maxframerate = 0;
				cap->flags = V4L2_FLAG_TUNER;
				ret = 0;
				goto radio_ioctl_done;
			}
		case VIDIOC_G_TUNER:
			{
				struct v4l2_tuner *t = arg;
				ret = v4l2_ioctl(id, cmd, arg);
				t->rangelow = (76*16);
				t->rangehigh = (108*16);
				t->signal = 0;
				goto radio_ioctl_done;
			}
		case VIDIOC_G_AUDIO:
		case VIDIOC_S_AUDIO:
		case VIDIOC_S_TUNER:
		case VIDIOC_S_FREQ:
		case VIDIOC_G_FREQ:
			{
				ret = v4l2_ioctl(id, cmd, arg);
				goto radio_ioctl_done;
			}
		case VIDIOC_QUERYCTRL:
		case VIDIOC_G_CTRL:
		case VIDIOC_S_CTRL:
			{
				ret = -EINVAL;
				goto radio_ioctl_done;
			}
		default:
			{
				ret = -ENOIOCTLCMD;
				goto radio_ioctl_done;
			}
	}
radio_ioctl_done:
	up(&btv->lock);
	return ret;
}

static int new_radio_device(struct bttv *btv)
{
	struct v4l2_device *v = &btv->radio_dev;
	sprintf(v->name, "Bttv2 radio device %d", btv->nr);
	v->type = V4L2_TYPE_RADIO;
	v->minor = radio[btv->nr];
	v->open = v4l2_radio_open;
	v->close = v4l2_radio_close;
	v->read = NULL;
	v->write = NULL;
	v->ioctl = v4l2_radio_ioctl;
	v->mmap = NULL;
	v->poll = NULL;
	v->initialize = v4l2_init_done;
	v->priv = btv;
	return v4l2_register_device(v);
}; 

/***************************************************************
 *                                                             *
 *                            VBI                              *
 *                                                             *
 ***************************************************************/

static int vbi_request_buffers(struct bttvc *btvc, struct v4l2_requestbuffers *req)
{
	struct bttvo *btvo = btvc->open;
	struct bttv *btv = btvo->dev;
	int max_count, i;
	
	if(btvc->req)
	{
		Dprintk(1,"bttv%d: buffers already requested!\n", btv->nr);
		return -EINVAL;
	}
	req->type=V4L2_BUF_TYPE_CAPTURE;
	if(req->count < 1)
		req->count = 1;
	/* FIXME calculate based on other opens too */
	max_count = (bttv_maxmem * 1024 * 1024) / btvo->vbi_bsize;
	if(req->count > max_count)
		req->count = max_count;
	Dprintk(1,"bttv%d: vbi granting %d buffers.\n", btv->nr, req->count);
	btvc->bufs = vmalloc_32(sizeof(struct stream_buffer)*req->count);
	if(!btvc->bufs)
	{
		printk("bttv%d: Could not vmalloc() capture buffers.\n", btv->nr);
		return -ENOMEM;
	}
	btvc->buf_vmalloc = sizeof(struct stream_buffer)*req->count;
	memset(btvc->bufs, 0, btvc->buf_vmalloc);
	for(i=0; i<req->count; i++)
	{
		btvc->bufs[i].buf.index = i;
		btvc->bufs[i].buf.type = req->type;
		btvc->bufs[i].buf.offset = i*PAGE_SIZE;
		btvc->bufs[i].buf.length = btvo->vbi_bsize;
		btvc->bufs[i].c = btvc;
	}
	btvc->req = req->count;
	return 0;
}

static void vbi_unrequest_buffers(struct bttvc *btvc)
{
	struct bttvo *btvo = btvc->open;
	struct bttv *btv = (struct bttv*)btvo->dev;
	int i, j;
	
	if(!btvc->req)
	{
		Dprintk(2,"bttv%d: caputre buffers notrequested!\n", btv->nr);
		return;
	}

	while (v4l2_q_del_head(&btvc->cap_done_q));
	while (v4l2_q_del_head(&btvc->cap_wait_q));
	
	if(!btvc->bufs)
	{
		printk("bttv%d: Hello... No buffers!\n", btv->nr);
		return;
	}

	for(i=0; i<btvc->req; i++)
	{
		if(btvc->bufs[i].hw.data)
		{
			for(j=0; j<btvo->vbi_bsize; j+=PAGE_SIZE)
				mem_map_unreserve(virt_to_page(btvc->bufs[i].hw.data+j));
			kfree(btvc->bufs[i].hw.data);
		}
	}
	vfree(btvc->bufs);
	btvc->bufs = NULL;
	btvc->buf_vmalloc = 0;
	btvc->req = 0;
	
	return;
}

static int vbi_queuebuffer(struct bttvc *btvc, struct v4l2_buffer *vbuf)
{
	struct bttvo *btvo = btvc->open;
	struct bttv *btv = btvo->dev;
	int i = vbuf->index;
	struct stream_buffer *buf = NULL;

	if (vbuf->type != V4L2_BUF_TYPE_CAPTURE)
	{
		Dprintk(1,"bttv%d: vbi QBUF wrong type\n", btv->nr);
		return -EINVAL;
	}
	if ((i < 0) || (i >= btvc->req))
	{
		Dprintk(1,"bttv%d: vbi QBUF buffer index %d is out of range\n", btv->nr, i);
		return -EINVAL;
	}
	buf = &btvc->bufs[i];
	if ((buf->buf.flags & V4L2_BUF_FLAG_QUEUED))
	{
		Dprintk(1,"bttv%d: vbi QBUF buffer %d is already queued\n", btv->nr, i);
		return -EINVAL;
	}
	if (!(buf->buf.flags & V4L2_BUF_FLAG_MAPPED))
	{
		Dprintk(1,"bttv%d: vbi QBUF buffer %d is not mapped\n", btv->nr, i);
		return -EINVAL;
	}
	buf->buf.flags &= ~V4L2_BUF_FLAG_DONE;
	buf->buf.flags |= V4L2_BUF_FLAG_QUEUED;
	v4l2_q_add_tail(&btvc->cap_wait_q, &buf->q);
	return 0;
}

static int vbi_dequeuebuffer(struct bttvc *btvc, struct v4l2_buffer *buf)
{
	struct bttvo *btvo = btvc->open;
	struct bttv *btv = (struct bttv*)btvo->dev;
	static struct stream_buffer *newbuf;

	if (buf->type != V4L2_BUF_TYPE_CAPTURE)
	{
		Dprintk(1,"bttv%d: vbi DQBUF wrong buffer type\n", btv->nr);
		return -EINVAL;
	}
	newbuf = v4l2_q_del_head(&btvc->cap_done_q);
	if (newbuf == NULL)
	{
		Dprintk(1,"bttv%d: vbi DQBUF nothing on done queue\n", btv->nr);
		return -EINVAL;
	}
	
	if(btvc ->start_time == 0)
		btvc->start_time = newbuf->buf.timestamp;
	btvc->int_time = newbuf->buf.timestamp;
	btvc->int_time -= btvc->start_time;
	btvo->perf.frames++;
	newbuf->buf.bytesused = (btvo->vbi_fmt.count[0]+btvo->vbi_fmt.count[1])*btv->vbi_bpl;
	btvo->perf.bytesout += newbuf->buf.bytesused;
	newbuf->buf.flags |= V4L2_BUF_FLAG_DONE | V4L2_BUF_FLAG_KEYFRAME;
	newbuf->buf.flags |= V4L2_BUF_FLAG_BOTFIELD|V4L2_BUF_FLAG_TOPFIELD;
	newbuf->buf.timestamp = btvc->int_time;
	if(btvc->last_frame>=0)
	{
		btvc->frame_no = v4l2_timestamp_correct(&newbuf->buf.timestamp, btv->tpf);
		newbuf->buf.timestamp += btvc->start_time;
		if(btvc->frame_no > btvc->last_frame+1)
		btvo->perf.framesdropped += btvc->frame_no - (btvc->last_frame + 1);
	} else
	{
		btvc->frame_no = 0;
		newbuf->buf.timestamp = btvc->start_time;
	}
	newbuf->buf.sequence = btvc->frame_no;
	btvc->last_frame = btvc->frame_no;
	
	newbuf->buf.flags &= ~V4L2_BUF_FLAG_QUEUED;
	memcpy(buf, &newbuf->buf, sizeof(struct v4l2_buffer));
	return 0;
}

static void vbi_streamoff(struct bttvc *btvc, __u32 type)
{
	struct bttvo *btvo = btvc->open;
	struct bttv *btv = btvo->dev;
	struct bttvc *c;
	unsigned long flags;
	int fst, i = 0;
	
/* all even/odd must be per open */

	Dprintk(2,"bttv%d: vbi streamoff\n", btv->nr);
	for(c=btv->streaming_vbis; c; c=c->next)
		if(c==btvc)
			break;
	if(c!=btvc)
		return;
	spin_lock_irqsave(&btv->s_lock, flags);
	btvc->enable = 0;
	spin_unlock_irqrestore(&btv->s_lock, flags);
	do
	{
		if((i++>3)||(dma_sync(btv)==0))
		{
			Dprintk(2,"bttv%d: RISC IRQ not responding - resetting DMA engine\n", btv->nr);
			bt848_stop_dma(btv);
			bt848_set_dma(btv);
			break;
		}
		spin_lock_irqsave(&btv->s_lock, flags);
		fst = ((btvc->wait_even)||(btvc->wait_odd));
		spin_unlock_irqrestore(&btv->s_lock, flags);
	} while(fst);
	spin_lock_irqsave(&btv->s_lock, flags);
	if(btv->streaming_vbis==btvc)
	{
		btv->streaming_vbis = btvc->next;
	} else
	{
		for(c=btv->streaming_vbis; c->next; c = c->next)
		{
			if(c->next==btvc) c->next = btvc->next;
			break;
		}
	}
	spin_unlock_irqrestore(&btv->s_lock, flags);
	Dprintk(2,"bttv%d: vbi streamoff done\n", btv->nr);
	return;
}

static int vbi_streamon(struct bttvc *btvc, __u32 type)
{
	struct bttvo *btvo = btvc->open;
	struct bttv *btv = btvo->dev;
	struct stream_buffer *buf;
	int i;
	unsigned long flags;
	struct bttvc *c;
	
	Dprintk(2,"bttv%d: vbi streamon\n", btv->nr);
	for(c=btv->streaming_vbis; c; c=c->next)
		if(c==btvc)
		{
			Dprintk(2,"bttv%d: vbi streaming already on!\n", btv->nr);
			return -EBUSY;
		}
	for(c=btv->streaming_vbis; c; c=c->next)
	{
		for(i=c->open->vbi_fmt.start[0]; i<(c->open->vbi_fmt.start[0]+c->open->vbi_fmt.count[0]); i++)
			if( (i>=btvo->vbi_fmt.start[0]) && 
			    (i<(btvo->vbi_fmt.start[0]+btvo->vbi_fmt.count[0])) )
			    	return -EBUSY;
		for(i=c->open->vbi_fmt.start[1]; i<(c->open->vbi_fmt.start[1]+c->open->vbi_fmt.count[1]); i++)
			if( (i>=btvo->vbi_fmt.start[1]) && 
			    (i<(btvo->vbi_fmt.start[1]+btvo->vbi_fmt.count[1])) )
			    	return -EBUSY;
	}
	
	if (type != V4L2_BUF_TYPE_CAPTURE)
	{
		Dprintk(1,"bttv%d: vbi wrong buffer type!\n", btv->nr);
		return -EINVAL;
	}
	while ((buf = v4l2_q_del_head(&btvc->cap_done_q)))
		buf->buf.flags &= ~V4L2_BUF_FLAG_QUEUED;
	memset(&btvo->perf, 0, sizeof(struct v4l2_performance));
	btvc->start_time = 0;
	btvc->last_frame = -1;
	btvc->wait_even = NULL;
	btvc->wait_odd = NULL;
	
	btvc->enable = 1;
	spin_lock_irqsave(&btv->s_lock, flags); 
	btvc->next = btv->streaming_vbis;
	btv->streaming_vbis = btvc;
	spin_unlock_irqrestore(&btv->s_lock, flags);
	
	Dprintk(2,"bttv%d: vbi streamon done\n", btv->nr);
	return 0;
}

static int v4l2_vbi_open(struct v4l2_device *v, int flags, void **idptr)
{
	struct bttv *btv = (struct bttv*)v->priv;
	int i;
	int ret;

	MOD_INC_USE_COUNT;
	down(&btv->lock);
	Dprintk(2,"bttv%d: VBI open for %s.\n", btv->nr, btv->vbi_dev.name);

	if(btv->radio)
	{
		ret = -EBUSY;
		MOD_DEC_USE_COUNT;
		goto vbi_open_done;
	}
	for (i = 0; i < MAX_VBI_OPENS; ++i)
		if (!btv->vbi_open_data[i].isopen)
			break;
	if (i == MAX_VBI_OPENS)/* No available open_data structures */
	{
		Dprintk(2,"bttv%d: No more opens on this device\n", btv->nr);
		ret = -EBUSY;
		MOD_DEC_USE_COUNT;
		goto vbi_open_done;
	}
	memset(&btv->vbi_open_data[i], 0, sizeof(struct bttvo));
	btv->vbi_open_data[i].isopen = 1;
	btv->vbi_open_data[i].index = i;
	btv->vbi_open_data[i].vbi = 1;
	btv->vbi_open_data[i].read.open = &(btv->vbi_open_data[i]);
	btv->vbi_open_data[i].stream.open = &(btv->vbi_open_data[i]);
	btv->vbi_open_data[i].input = btv->input;
	btv->vbi_open_data[i].dev = v->priv;
	*idptr = &btv->vbi_open_data[i];
	if (flags & O_NONCAP) /*  Non-capturing open */
	{
		btv->vbi_open_data[i].noncapturing = 1;
	} else
	{
		btv->vbi_open_data[i].noncapturing = 0;
		btv->vbi_open_data[i].vbi_fmt.sampling_rate = 0; /* invalidate format */
		v4l2_q_init(&btv->vbi_open_data[i].read.cap_wait_q);
		v4l2_q_init(&btv->vbi_open_data[i].read.cap_done_q);
		v4l2_q_init(&btv->vbi_open_data[i].stream.cap_wait_q);
		v4l2_q_init(&btv->vbi_open_data[i].stream.cap_done_q);
		init_waitqueue_head(&btv->vbi_open_data[i].waitq);
	}
	
	bt848_open(btv, 1);
	Dprintk(2,"bttv%d: VBI open succeeded\n", btv->nr);
	
	ret = 0;
vbi_open_done:
	up(&btv->lock);
	return ret;
}

static void v4l2_vbi_close(void *id)
{
	struct bttvo *btvo = (struct bttvo *)id;
	struct bttv *btv = (struct bttv*)btvo->dev;
	
	down(&btv->lock);

	Dprintk(2,"bttv%d: VBI close for %s.\n", btv->nr, btv->vbi_dev.name);
	
	if(!btvo->noncapturing)
	{
		/* shut off vbi capture here */
		vbi_streamoff(&btvo->read, V4L2_BUF_TYPE_CAPTURE);
		vbi_unrequest_buffers(&btvo->read);
		vbi_streamoff(&btvo->stream, V4L2_BUF_TYPE_CAPTURE);
		vbi_unrequest_buffers(&btvo->stream);
		Dprintk(2,"bttv%d: VBI capture stopped.\n", btv->nr);
	}
	
	btvo->isopen = 0;
	bt848_close(btv, 0);

	up(&btv->lock);
	MOD_DEC_USE_COUNT;
}

static int v4l2_vbi_ioctl(void *id, unsigned int cmd, void *arg)
{
	struct bttvo *btvo = (struct bttvo *)id;
	struct bttv *btv = (struct bttv *)btvo->dev;
	int ret;
	
	down(&btv->lock);
	switch(cmd)
	{
		case VIDIOC_QUERYCAP:
			{
				struct v4l2_capability *cap = arg;
				sprintf(cap->name, "Bttv2(%d) - VBI", btv->nr);
				cap->type = V4L2_TYPE_VBI;
				cap->inputs = bttv_tvcards[btv->type].video_inputs;
				cap->outputs = 0;
				cap->audios = 0;
				cap->maxwidth = btv->vbi_bpl;
				cap->maxheight = tvnorms[btv->norm].vdelay/2;
				cap->minwidth = cap->maxwidth;
				cap->minheight = 1;
				cap->maxframerate = 30;
				cap->flags = V4L2_FLAG_TUNER|V4L2_FLAG_STREAMING|V4L2_FLAG_SELECT|V4L2_FLAG_READ;
				ret = 0;
				goto vbi_ioctl_done;
			}
		case VIDIOC_ENUMSTD:
		case VIDIOC_G_STD:
		case VIDIOC_S_STD:
		case VIDIOC_G_TUNER:
		case VIDIOC_S_TUNER:
		case VIDIOC_ENUMINPUT:
		case VIDIOC_G_INPUT:
		case VIDIOC_S_INPUT:
		case VIDIOC_G_AUDIO:
		case VIDIOC_S_AUDIO:
		case VIDIOC_G_FREQ:
		case VIDIOC_S_FREQ:
		case VIDIOC_QUERYMENU:
		case VIDIOC_QUERYCTRL:
		case VIDIOC_G_CTRL:
		case VIDIOC_S_CTRL:
		case VIDIOC_G_PARM:
		case VIDIOC_S_PARM:
		case VIDIOC_G_PERF:
			{
				up(&btv->lock);
				return v4l2_ioctl(id, cmd, arg);
			}
		case VIDIOC_ENUM_PIXFMT:
			{
				struct v4l2_fmtdesc *f = arg;
				if(f->index!=0)
				{
					ret = -EINVAL;
					goto vbi_ioctl_done;
				}
				sprintf(f->description, "Unsigned Byte");
				f->pixelformat = V4L2_VBI_SF_UBYTE;
				f->flags = 0;
				f->depth = 8;
				ret = 0;
				goto vbi_ioctl_done;
			}
		case VIDIOC_S_FMT:
			{
				struct v4l2_format *f = arg;
				struct bttvc *btvc;
				int maxlines, i;
				if(btvo->noncapturing)
				{
					Dprintk(1,"bttv%d: S_FMT illegal on noncap open.\n", btv->nr);
					ret = -EPERM;
					goto vbi_ioctl_done;
				}
				for(btvc = btv->streaming_vbis; btvc; btvc = btvc->next)
					if(btvc==&btvo->stream)
					{
						Dprintk(1,"bttv%d: S_FMT while streaming.\n", btv->nr);
						ret = -EBUSY;
						goto vbi_ioctl_done;
					}
/*				if(f->type!=V4L2_BUF_TYPE_CAPTURE)
				{
					ret = -EINVAL;
					goto vbi_ioctl_done;
				} */
				maxlines = (tvnorms[btv->norm].sheight==576)?625:525;
				if( (f->fmt.vbi.count[0]<0)
					|| (f->fmt.vbi.start[0]<0)
					|| ((f->fmt.vbi.start[0]+f->fmt.vbi.count[0])>(tvnorms[btv->norm].vdelay/2)) )
				{
					Dprintk(1, "bttv%d: vbi invalid odd s_fmt %d:%d (%d,%d)\n", btv->nr, f->fmt.vbi.start[0], f->fmt.vbi.count[0], tvnorms[btv->norm].vdelay/2, maxlines);
					ret = -EINVAL;
					goto vbi_ioctl_done;
				}
				if( (f->fmt.vbi.count[1]<0)
					|| (f->fmt.vbi.start[1]<((maxlines/2)+1))
					|| ((f->fmt.vbi.start[1]+f->fmt.vbi.count[1])>((maxlines/2)+1+(tvnorms[btv->norm].vdelay/2))) )
				{
					Dprintk(1, "bttv%d: vbi invalid even s_fmt %d:%d (%d,%d)\n", btv->nr, f->fmt.vbi.start[1], f->fmt.vbi.count[1], tvnorms[btv->norm].vdelay/2, maxlines);
					ret = -EINVAL;
					goto vbi_ioctl_done;
				}
				i = 4096;
				while(i<((f->fmt.vbi.count[0]+f->fmt.vbi.count[1])*btv->vbi_bpl))
					i*=2;
				if((btvo->stream.map)&&(i>btvo->vbi_bsize))
				{
					Dprintk(1, "bttv%d: vbi format too big for mapped buffers\n", btv->nr);
					ret = -EINVAL;
					goto vbi_ioctl_done;
				}
				f->fmt.vbi.sampling_rate = tvnorms[btv->norm].Fsc;
				f->fmt.vbi.offset = 0;
				f->fmt.vbi.samples_per_line = btv->vbi_bpl;
				f->fmt.vbi.sample_format = V4L2_VBI_SF_UBYTE;
				f->fmt.vbi.flags = 0;
				memcpy(&btvo->vbi_fmt, &f->fmt.vbi, sizeof(struct v4l2_vbi_format));
				btvo->vbi_fmt.start[1] -= (maxlines/2)+1;
				btvo->vbi_bsize = i;
				Dprintk(1, "bttv%d: vbi s_fmt %d:%d %d:%d\n", btv->nr, btvo->vbi_fmt.start[0], btvo->vbi_fmt.count[0], btvo->vbi_fmt.start[1], btvo->vbi_fmt.count[1]);
				ret = 0;
				goto vbi_ioctl_done;
			}
		case VIDIOC_G_FMT:
			{
				struct v4l2_format *f = arg;
				int maxlines;
/*				if(f->type!=V4L2_BUF_TYPE_CAPTURE)
				{
					ret = -EINVAL;
					goto vbi_ioctl_done;
				} */
				if(btvo->vbi_fmt.sampling_rate==0)
				{
					ret = -EINVAL;
					goto vbi_ioctl_done;
				}
				maxlines = (tvnorms[btv->norm].sheight==768)?625:525;
				memcpy(&f->fmt.vbi, &btvo->vbi_fmt, sizeof(struct v4l2_vbi_format));
				f->fmt.vbi.start[1] += (maxlines/2)+1;
				ret = 0;
				goto vbi_ioctl_done;
			}
		case VIDIOC_REQBUFS:
			{
				struct v4l2_requestbuffers *req = arg;
				if(btvo->noncapturing)
				{
					Dprintk(1,"bttv%d: REQBUFS illegal on noncap open.\n", btv->nr);
					ret = -EPERM;
					goto vbi_ioctl_done;
				}
				if(btvo->vbi_fmt.sampling_rate==0)
				{
					ret = -EINVAL;
					goto vbi_ioctl_done;
				}
				ret = vbi_request_buffers(&btvo->stream, req);
				goto vbi_ioctl_done;
			}
		case VIDIOC_QUERYBUF:
			{
				struct v4l2_buffer *b = arg;
				
				if (btvo->noncapturing)
				{
					Dprintk(1,"bttv%d: vbi querybuf illegal on noncap open.\n", btv->nr);
					ret = -EPERM;
					goto vbi_ioctl_done;
				}
				if((b->index<0)||(b->index>=btvo->stream.req))
				{
					ret = -EINVAL;
					goto vbi_ioctl_done;
				}
				if((b->type & V4L2_BUF_TYPE_field)!=(btvo->stream.bufs[b->index].buf.type & V4L2_BUF_TYPE_field))
				{
					ret = -EINVAL;
					goto vbi_ioctl_done;
				}
				memcpy(b, &btvo->stream.bufs[b->index].buf, sizeof(struct v4l2_buffer));
				Dprintk(2,"bttv%d: vbi querybuf %d: %d %d\n", btv->nr, b->index, b->length, b->offset);
				ret = 0;
				goto vbi_ioctl_done;
			}
		case VIDIOC_QBUF:
			{
				struct v4l2_buffer *buf = arg;
				if (btvo->noncapturing)
				{
					Dprintk(1,"bttv%d: vbi QBUF illegal in non-capturing open\n", btv->nr);
					ret = -EPERM;
					goto vbi_ioctl_done;
				}
				ret = vbi_queuebuffer(&btvo->stream, buf);
				goto vbi_ioctl_done;
			}
		case VIDIOC_DQBUF:
			{
				struct v4l2_buffer *buf = arg;
				if (btvo->noncapturing)
				{
					Dprintk(1,"bttv%d: vbi DQBUF illegal in non-capturing open\n", btv->nr);
					ret = -EPERM;
					goto vbi_ioctl_done;
				}
				ret = vbi_dequeuebuffer(&btvo->stream, buf);
				goto vbi_ioctl_done;
			}
		case VIDIOC_STREAMON:
			{
				__u32 *type = arg;
				if (btvo->noncapturing)
				{
					Dprintk(1,"btv%d: vbi STREAMON illegal in non-capturing open\n", btv->nr);
					ret = -EPERM;
					goto vbi_ioctl_done;
				}
				if (btvo->stream.map!=btvo->stream.req)
				{
					Dprintk(1,"btv%d: vbi no all buffers mapped!\n", btv->nr);
					ret = -EINVAL;
					goto vbi_ioctl_done;
				}
				ret = vbi_streamon(&btvo->stream, *type);
				goto vbi_ioctl_done;
			}
		case VIDIOC_STREAMOFF:
			{
				__u32	*type = arg;
				if (btvo->noncapturing)
				{
					Dprintk(1,"bttv%d: vbi STREAMOFF illegal in non-capturing open\n", btv->nr);
					ret = -EPERM;
					goto vbi_ioctl_done;
				}
				vbi_streamoff(&btvo->stream, *type);
				ret = 0;
				goto vbi_ioctl_done;
			}
		
		case VIDIOC_ENUM_FBUFFMT:
		case VIDIOC_G_FBUF:
		case VIDIOC_S_FBUF:
		case VIDIOC_G_WIN:
		case VIDIOC_S_WIN:
		case VIDIOC_PREVIEW:
			{
				ret = -EINVAL;
				goto vbi_ioctl_done;
			}
		
		default:
			{
				ret = -ENOIOCTLCMD;
				goto vbi_ioctl_done;
			}
	}
vbi_ioctl_done:
	up(&btv->lock);
	return ret;
}

/* MMAP */

static void vbi_mmap_vma_open(struct vm_area_struct *vma)
{
	struct bttv *btv = (struct bttv*)v4l2_device_from_file(vma->vm_file);
	struct bttvo *btvo = (struct bttvo*)v4l2_openid_from_file(vma->vm_file);
	struct bttvc *btvc = &btvo->stream;
 	struct stream_buffer *buf;

	if ((!btv)||(!btvo))
		return;
		
#if  LINUX_VERSION_CODE > KERNEL_VERSION(2,3,0)
	buf = mmap_stream_buffer_from_offset(btvc, vma->vm_pgoff * PAGE_SIZE);
#else
	buf = mmap_stream_buffer_from_offset(btvc, vma->vm_offset);
#endif
	if(!buf)
		return;
	buf->hw.vma_nopen++;
	Dprintk(2,"bttv%d: vma_open called on buffer %d (refcount %d)\n",
		btv->nr, buf->buf.index, buf->hw.vma_nopen);
}


static void vbi_mmap_vma_unmap(struct vm_area_struct *vma, unsigned long address, size_t size)
{
	struct bttv *btv = (struct bttv*)v4l2_device_from_file(vma->vm_file);
	struct bttvo *btvo = (struct bttvo*)v4l2_openid_from_file(vma->vm_file);
	struct bttvc *btvc = &btvo->stream;
	struct bttvc *c;
 	struct stream_buffer *buf;
	
	if((!btv)||(!btvo))
		return;
#if  LINUX_VERSION_CODE > KERNEL_VERSION(2,3,0)
	buf = mmap_stream_buffer_from_offset(btvc, vma->vm_pgoff * PAGE_SIZE);
#else
	buf = mmap_stream_buffer_from_offset(btvc, vma->vm_offset);
#endif
	if(!buf)
	{
		Dprintk(1,"bttv%d: vbi munmap: invalid buffer!\n", btv->nr);
		return;
	}
	buf->hw.vma_nopen--;
	Dprintk(1,"bttv%d: vbi_vma_unmap called on buffer %d (refcount %d)\n",
		btv->nr, buf->buf.index, buf->hw.vma_nopen);
	if (buf->hw.vma_nopen > 0)
		return;

	for(c = btv->streaming_vbis; c; c=c->next)
		if(c==btvc)
		{
			Dprintk(1,"bttv%d: Warning - final munmap() called while streaming\n", btv->nr);
			vbi_streamoff(btvc, V4L2_BUF_TYPE_CAPTURE);
		}

	v4l2_q_yank_node(&btvc->cap_wait_q, &buf->q);
	v4l2_q_yank_node(&btvc->cap_done_q, &buf->q);
	if (btvc->map > 0)
		--btvc->map;
	if(btvc->map<=0)
		vbi_unrequest_buffers(btvc);
}

static struct vm_operations_struct vbi_vma_operations =
{
	open:	vbi_mmap_vma_open,
	unmap:	vbi_mmap_vma_unmap,
};

static int v4l2_vbi_mmap(void *id, struct vm_area_struct *vma)
{
	struct bttvo *btvo = (struct bttvo *)id;
	struct bttv *btv = btvo->dev;
	struct bttvc *btvc = &btvo->stream;
	struct stream_buffer *buf;
	int ret, i;

	down(&btv->lock);
	
	Dprintk(2,"bttv%d: vbi mmap() for %s.\n", btv->nr, btv->video_dev.name);

	if (btvo->noncapturing)
	{
		Dprintk(1,"bttv%d: vbi mmap() called on non-capturing open\n", btv->nr);
		ret = -ENODEV;
		goto vbi_mmap_done;
	}

#if  LINUX_VERSION_CODE > KERNEL_VERSION(2,3,0)
	buf = mmap_stream_buffer_from_offset(btvc, vma->vm_pgoff * PAGE_SIZE);
#else
	buf = mmap_stream_buffer_from_offset(btvc, vma->vm_offset);
#endif
	if (buf == NULL)
	{
		Dprintk(1,"bttv%d: mmap() Invalid offset parameter\n", btv->nr);
		ret = -EINVAL;/* no such buffer */
		goto vbi_mmap_done;
	}
	if ((unsigned long)btvo->vbi_bsize != (unsigned long)(vma->vm_end-vma->vm_start))
	{
		Dprintk(1,"bttv%d: vbi mmap() Wrong length parameter %u %lu\n", btv->nr, btvo->cap_bsize, vma->vm_end - vma->vm_start);
		ret = -EINVAL;/* wrong length */
		goto vbi_mmap_done;
	}
	if(!buf->hw.data) /* If we have not yet malloc'ed data, do it now */
	{
		Dprintk(2,"bttv%d: vbi kmalloc() buffers\n", btv->nr); 
		buf->hw.data = kmalloc(btvo->vbi_bsize, GFP_KERNEL);
		if(!buf->hw.data)
		{
			printk("bttv%d: vbi mmap() failed to kmalloc() buffer\n", btv->nr);
			ret = -ENOMEM;
			goto vbi_mmap_done;
		}
		for(i=0; i<btvo->vbi_bsize; i+=PAGE_SIZE)
			mem_map_reserve(virt_to_page(buf->hw.data+i));
		memset(buf->hw.data, 0, btvo->vbi_bsize);
	}

	if (buf->buf.flags & V4L2_BUF_FLAG_MAPPED)
	{
		Dprintk(1,"bttv%d: vbi mmap() Buffer is already mapped\n", btv->nr);
		ret = -EINVAL;/* already mapped */
		goto vbi_mmap_done;
	}
	if (remap_page_range(vma->vm_start, virt_to_phys(buf->hw.data), btvo->vbi_bsize, vma->vm_page_prot))
		return -EAGAIN;
	buf->buf.flags |= V4L2_BUF_FLAG_MAPPED;
	++btvc->map;
	vma->vm_ops = &vbi_vma_operations;
	if (vma->vm_ops->open)
		vma->vm_ops->open(vma);

	ret = 0;

vbi_mmap_done:	
	Dprintk(2,"bttv%d: vbi mmap() %s: %d\n", btv->nr, btv->video_dev.name, ret);
	up(&btv->lock);
	return ret;
}

static long v4l2_vbi_read(void *id, char *buf, unsigned long len, int noblock)
{
	struct bttvo *btvo = (struct bttvo *)id;
	struct bttv *btv = btvo->dev;
	struct bttvc *btvc = &btvo->read;
	long my_timeout;
	int ret;
	int i;
	DECLARE_WAITQUEUE(wait, current);
	
	down(&btv->lock);
	
	Dprintk(2,"bttv%d: vbi read() for %s.\n", btv->nr, btv->video_dev.name);

	if(btvo->noncapturing)
	{
		Dprintk(1,"bttv%d: vbi read(): illegal on noncap open.\n", btv->nr);
		ret = -EINVAL;
		goto vbi_read_done;
	}
	if(!btvo->reading)
	{
		struct v4l2_requestbuffers req;
		
		if(btvc->req)
		{
			Dprintk(1,"bttv%d: vbi read(): capture buffers already requested.\n", btv->nr);
			ret = -EBUSY;
			goto vbi_read_done;
		}
		req.type = V4L2_BUF_TYPE_CAPTURE;
		req.count = 1;
		ret = vbi_request_buffers(btvc, &req);
		if(ret!=0)
		{
			printk("bttv%d: vbi read(): could not request buffers.\n", btv->nr);
			goto vbi_read_done;
		}
		btvc->bufs[0].hw.data = kmalloc(btvo->vbi_bsize, GFP_KERNEL);
		if(!btvc->bufs[0].hw.data)
		{
			printk("bttv%d: vbi read(): could not allocate capture memory.\n", btv->nr);
			vbi_unrequest_buffers(btvc);
			ret = -ENOMEM;
			goto vbi_read_done;
		}
		btvo->read_buffer.type = V4L2_BUF_TYPE_CAPTURE;
		btvo->read_buffer.index = 0;
		btvc->bufs[0].buf.flags |= V4L2_BUF_FLAG_MAPPED;
		vbi_queuebuffer(btvc, &btvo->read_buffer);
		vbi_streamon(btvc, V4L2_BUF_TYPE_CAPTURE);
		btvo->reading = 1;
	}
	if(btvo->reading) // read for this device
	{
		btvo->read_buffer.type = V4L2_BUF_TYPE_CAPTURE;
		i=vbi_dequeuebuffer(btvc, &btvo->read_buffer);
		if((i<0)&&(noblock))
		{
			ret = -EWOULDBLOCK;
			goto vbi_read_done;
		}
		my_timeout = HZ;
		add_wait_queue(&btvo->waitq, &wait);
		current->state = TASK_INTERRUPTIBLE;
		do
		{
			up(&btv->lock);
			my_timeout = schedule_timeout(my_timeout);
			if(signal_pending(current))
			{
				remove_wait_queue(&btvo->waitq, &wait);
				current->state = TASK_RUNNING;
				ret = -EINTR;
				goto vbi_read_done;
			}
			if(my_timeout<=0)
			{
				remove_wait_queue(&btvo->waitq, &wait);
				current->state = TASK_RUNNING;
				Dprintk(1,"bttv%d: Timeout on read().\n", btv->nr);
				ret = 0;
				goto vbi_read_done;
			}
			down(&btv->lock);
			i=vbi_dequeuebuffer(btvc, &btvo->read_buffer);
		} while(i<0);
		remove_wait_queue(&btvo->waitq, &wait);
		current->state = TASK_RUNNING;
		if(len < btvo->read_buffer.bytesused)
		{
			ret = -EINVAL;
			goto vbi_read_done;
		}
		copy_to_user(buf, btvc->bufs[0].hw.data, btvo->read_buffer.bytesused);
		ret = btvo->read_buffer.bytesused;
		vbi_streamoff(btvc, V4L2_BUF_TYPE_CAPTURE);
		vbi_unrequest_buffers(btvc);
		btvo->reading = 0;
		goto vbi_read_done;
	}
	ret = -EBUSY;
vbi_read_done:
	up(&btv->lock);
	return ret;
}

static int new_vbi_device(struct bttv *btv)
{
	struct v4l2_device *v = &btv->vbi_dev;
	sprintf(v->name, "Bttv2 VBI device %d", btv->nr);
	v->type = V4L2_TYPE_VBI;
	v->minor = bttv_vbi[btv->nr];
	v->open = v4l2_vbi_open;
	v->close = v4l2_vbi_close;
	v->read = v4l2_vbi_read;
	v->write = NULL;
	v->ioctl = v4l2_vbi_ioctl;
	v->mmap = v4l2_vbi_mmap;
	v->poll = v4l2_poll;
	v->initialize = v4l2_init_done;
	v->priv = btv;
	return v4l2_register_device(v);
}; 

/* Initialisation stuff */  

static void __devinit handle_chipset(void)
{
	struct pci_dev *dev = NULL;
	
	/*	Just in case some nut set this to something dangerous */
	if (triton1)
		triton1=BT848_INT_ETBF;
	
	while ((dev = pci_find_device(PCI_VENDOR_ID_SI, PCI_DEVICE_ID_SI_496, dev))) 
	{
		/* Beware the SiS 85C496 my friend - rev 49 don't work with a bttv */
		printk(KERN_WARNING "BT848 and SIS 85C496 chipset don't always work together.\n");
	}			

        while ((dev = pci_find_device(PCI_VENDOR_ID_INTEL,
				      PCI_DEVICE_ID_INTEL_82441, dev))) 
        {
                unsigned char b;
                pci_read_config_byte(dev, 0x53, &b);
                Dprintk(1,KERN_INFO "bttv: Host bridge: 82441FX Natoma, ");
                Dprintk(1,"bufcon=0x%02x\n",b);
        }

	while ((dev = pci_find_device(PCI_VENDOR_ID_INTEL, PCI_DEVICE_ID_INTEL_82437, dev))) 
	{
		Vprintk(1,KERN_INFO "bttv: Host bridge 82437FX Triton PIIX\n");
		triton1=BT848_INT_ETBF;
	}
	
}

static int __devinit init_bt848(struct bttv *btv)
{
        init_MUTEX(&btv->lock);

	/* dump current state of the gpio registers before changing them,
	 * might help to make a new card work */
	if (bttv_gpio)
		bttv_gpio_tracking(btv,"init #1");
		
	/* reset the bt848 */
	btwrite(0, BT848_SRESET);
	Dprintk(1,KERN_DEBUG "bttv%d: bt848_mem: 0x%lx\n", btv->nr, (unsigned long) btv->bt848_mem);

/*	btwrite(0, BT848_TDEC); */
        btwrite(0x10, BT848_COLOR_CTL);
	btwrite(0x00, BT848_CAP_CTL);
	/* set planar and packed mode trigger points and         */
	/* set rising edge of inverted GPINTR pin as irq trigger */
	btwrite(BT848_GPIO_DMA_CTL_PKTP_32|
		BT848_GPIO_DMA_CTL_PLTP1_16|
		BT848_GPIO_DMA_CTL_PLTP23_16|
		BT848_GPIO_DMA_CTL_GPINTC|
		BT848_GPIO_DMA_CTL_GPINTI, 
		BT848_GPIO_DMA_CTL);

        /* select direct input */
	btwrite(0x00, BT848_GPIO_REG_INP);
	btwrite(0x00, BT848_GPIO_OUT_EN);
	if (bttv_gpio)
		bttv_gpio_tracking(btv,"init #2");

	btwrite(BT848_IFORM_MUX1 | BT848_IFORM_XTAUTO | BT848_IFORM_AUTO,
		BT848_IFORM);

	btwrite(0xd8, BT848_CONTRAST_LO);

	btwrite(0x20, BT848_E_VSCALE_HI);
	btwrite(0x20, BT848_O_VSCALE_HI);
	btwrite(/*BT848_ADC_SYNC_T|*/
		BT848_ADC_RESERVED|BT848_ADC_CRUSH, BT848_ADC);

	btwrite(BT848_CONTROL_LDEC, BT848_E_CONTROL);
	btwrite(BT848_CONTROL_LDEC, BT848_O_CONTROL);

	btwrite(0x00, BT848_E_SCLOOP);
	btwrite(0x00, BT848_O_SCLOOP);

	/* clear interrupt status */
	btwrite(0xfffffUL, BT848_INT_STAT);
        
	/* set interrupt mask */
	btwrite(btv->triton1|
                /*BT848_INT_PABORT|BT848_INT_RIPERR|BT848_INT_PPERR|
                  BT848_INT_FDSR|BT848_INT_FTRGT|BT848_INT_FBUS|
                BT848_INT_VSYNC|*/
		BT848_INT_GPINT|
		BT848_INT_SCERR|
		BT848_INT_RISCI|BT848_INT_OCERR|BT848_INT_VPRES|
		BT848_INT_FMTCHG|BT848_INT_HLOCK,
		BT848_INT_MASK);
	
	/* needs to be done before i2c is registered */
	if (btv->type == BTTV_HAUPPAUGE || btv->type == BTTV_HAUPPAUGE878)
		bttv_hauppauge_boot_msp34xx(btv);
		
	/* register i2c */
	btv->tuner_type=-1;
	init_bttv_i2c(btv);
	
	/* some card-specific stuff (needs working i2c) */
	bttv_init_card(btv);

	/*
	 *	Now add the template and register the device unit.
	 */
        new_video_device(btv);
        new_vbi_device(btv);
        if(radio[btv->nr]>0)
	        new_radio_device(btv);

// *************** Set up all defaults here
	bt848_muxsel(btv, 1); // SET DEFAULT INPUT HERE
	btv->norm = bttv_norm[btv->nr];
	bt848_set_standard(btv); // SET DEFAULT STANDARD HERE
	default_controls(btv); // default all controls

	return 0;
}

/* ----------------------------------------------------------------------- */

static char *irq_name[] = { "FMTCHG", "VSYNC", "HSYNC", "OFLOW", "HLOCK",
				"VPRES", "6", "7", "I2CDONE", "GPINT", "10",
				"RISCI", "FBUS", "FTRGT", "FDSR", "PPERR",
				"RIPERR", "PABORT", "OCERR", "SCERR" };

static void bttv_irq(int irq, void *dev_id, struct pt_regs * regs)
{
	u32 stat,astat;
	u32 dstat;
	int count;
	struct bttv *btv;
	int odd;

	btv=(struct bttv *)dev_id;
	count=0;
	while (1) 
	{
		/* get/clear interrupt status bits */
		stat=btread(BT848_INT_STAT);
		astat=stat&btread(BT848_INT_MASK);
		if (!astat)
		{
			return;
		}
		btwrite(stat,BT848_INT_STAT);

		/* get device status bits */
		dstat=btread(BT848_DSTATUS);
		
		if (irq_debug) {
			int i;
			printk(KERN_DEBUG "bttv%d: irq loop=%d risc=%x, bits:",
			       btv->nr, count, stat>>28);
			for (i = 0; i < (sizeof(irq_name)/sizeof(char*)); i++) {
				if (stat & (1 << i))
					printk(" %s",irq_name[i]);
				if (astat & (1 << i))
					printk("*");
			}
			if (stat & BT848_INT_HLOCK)
				printk("   HLOC => %s", (dstat & BT848_DSTATUS_HLOC)
					? "yes" : "no");
			if (stat & BT848_INT_VPRES)
				printk("   PRES => %s", (dstat & BT848_DSTATUS_PRES)
					? "yes" : "no");
			if (stat & BT848_INT_FMTCHG)
				printk("   NUML => %s", (dstat & BT848_DSTATUS_PRES)
					? "625" : "525");
			printk("\n");
 		}

		if (astat&BT848_INT_GPINT)
			wake_up_interruptible(&btv->gpioq);
	
    		if (astat&(BT848_INT_SCERR|BT848_INT_OCERR)) {
			if (bttv_verbose)
				printk("bttv%d: irq:%s%s risc_count=%08x\n",
				       btv->nr,
				       (astat&BT848_INT_SCERR) ? " SCERR" : "",
				       (astat&BT848_INT_OCERR) ? " OCERR" : "",
				       btread(BT848_RISC_COUNT));
			// btv->errors++;
			if (btv->errors < BTTV_ERRORS) {
				spin_lock(&btv->s_lock);
				// ***************
				// stop dma
				// set start add to start of vbi
				// restart dma
				spin_unlock(&btv->s_lock);
			} else {
				if (bttv_verbose)
					printk("bttv%d: aiee: error loops\n",btv->nr);
				// bt848_offline(btv);
			}
		}
		spin_lock(&btv->s_lock);
		odd = (stat&(1<<28)) ? 0:1;
		if ((astat&BT848_INT_RISCI)&&(odd!=btv->last_irq))
		{
			struct stream_buffer *c_done = NULL;
			int i=0;
			struct bttvc *c;
			char * vaddr;
			
			btv->last_irq = odd;
			Dprintk(3,"bttv%d: IRQ_RISCI %d\n", btv->nr, odd);
			if(odd) /* ODD */
			{
				/* Find any completed buffers for later processing */
				for(c = btv->streaming_caps; c; c = c->next)
				{
					if((c->wait_odd)&&
					   (!(c->open->cap_fmt.flags & V4L2_FMT_FLAG_INTERLACED)))
					{
						c->wait_odd->next = c_done;
						c_done = c->wait_odd;
					}
					c->wait_odd = NULL;
					if(c->enable)
					{	
						if(c->open->cap_fmt.flags&(V4L2_FMT_FLAG_INTERLACED|V4L2_FMT_FLAG_TOPFIELD))
						{
							c->wait_odd = v4l2_q_del_head(&c->cap_wait_q);
							if(c->wait_odd)
								*btv->cap_odd_jmp = c->wait_odd->hw.risc_odd_bus;
							else
								*btv->cap_odd_jmp = c->open->cap_dummy.risc_odd_bus;
							i++;
						}
						if(btv->o_stat!=c->open->index)
						{
							bt848_set_geo(btv, &c->open->cap_geo, 1);
							btv->o_stat=c->open->index;
						}
					}
				}
				if(i==0) /* no captures */
				{
					if(btv->pv)
					{
						if(btv->o_stat!=STAT_PV)
						{
							*btv->cap_odd_jmp = btv->pv_buf.risc_odd_bus;
							bt848_set_geo(btv, &btv->pv_geo, 1);
							btv->o_stat = STAT_PV;
						}
					} else
					{
						if(btv->o_stat!=STAT_DUMMY)
						{
							*btv->cap_odd_jmp = btv->dummy_buf.risc_odd_bus;
							bt848_set_geo(btv, &btv->dummy_geo, 1);
							btv->o_stat = STAT_DUMMY;
						}
					}
				}
			} else /* EVEN */
			{
				for(c = btv->streaming_caps; c; c = c->next)
				{
					if(c->wait_even)
					{
						c->wait_even->next = c_done;
						c_done = c->wait_even;
					}
					c->wait_even = NULL;
					if( (c->enable)||
					    ( (c->open->cap_fmt.flags&V4L2_FMT_FLAG_INTERLACED)&&
					      (c->wait_odd) ) )
					{
						if(c->open->cap_fmt.flags&(V4L2_FMT_FLAG_BOTFIELD|V4L2_FMT_FLAG_INTERLACED))
						{
							if(c->open->cap_fmt.flags&V4L2_FMT_FLAG_BOTFIELD)
								c->wait_even = v4l2_q_del_head(&c->cap_wait_q);
							else
								c->wait_even = c->wait_odd;
							if(c->wait_even)
								*btv->cap_even_jmp = c->wait_even->hw.risc_even_bus;
							else
								*btv->cap_even_jmp = c->open->cap_dummy.risc_even_bus;
							i++;
						}
						if(btv->e_stat!=c->open->index)
						{
							bt848_set_geo(btv, &c->open->cap_geo, 0);
							btv->e_stat=c->open->index;
						}
					}
				}
				if(i==0)
				{
					if (btv->pv)
					{
						if(btv->e_stat!=STAT_PV)
						{
							*btv->cap_even_jmp = btv->pv_buf.risc_even_bus;
							bt848_set_geo(btv, &btv->pv_geo, 0);
							btv->e_stat = STAT_PV;
						}
					} else
					{
						if(btv->e_stat!=STAT_DUMMY)
						{
							*btv->cap_even_jmp = btv->dummy_buf.risc_even_bus;
							bt848_set_geo(btv, &btv->dummy_geo, 0);
							btv->e_stat = STAT_DUMMY;
						}
					}
				}
			}
			/* VBI */
			if(btv->vbi_opens>0)
			{
				if(odd) /* ODD */
				{
					for(c = btv->streaming_vbis; c; c = c->next)
					{
						if(c->enable)
							c->wait_odd = v4l2_q_del_head(&c->cap_wait_q);
						else
							c->wait_odd = NULL;
						if(c->wait_odd)
						{
							vaddr = c->wait_odd->hw.data;
							for(i=c->open->vbi_fmt.start[0]; i<(c->open->vbi_fmt.start[0]+c->open->vbi_fmt.count[0]); i++)
							{
								btv->vbi_odd_tab[i*2] = cpu_to_le32(virt_to_bus(vaddr));
								vaddr+=btv->vbi_bpl;
							}
						} else
						{
							for(i=c->open->vbi_fmt.start[0]; i<(c->open->vbi_fmt.start[0]+c->open->vbi_fmt.count[0]); i++)
								btv->vbi_odd_tab[i*2] = btv->vbi_dummy_phys;
						}
					}
				} else /* EVEN */
				{
					for(c = btv->streaming_vbis; c; c = c->next)
					{
						if(c->wait_even)
						{
							c->wait_even->next = c_done;
							c_done = c->wait_even;
						}
						c->wait_even = c->wait_odd;
						if(c->wait_even)
						{
							vaddr = c->wait_even->hw.data + (c->open->vbi_fmt.count[0]*btv->vbi_bpl);
							for(i=c->open->vbi_fmt.start[1]; i<(c->open->vbi_fmt.start[1]+c->open->vbi_fmt.count[1]); i++)
							{
								btv->vbi_even_tab[i*2] = cpu_to_le32(virt_to_bus(vaddr));
								vaddr+=btv->vbi_bpl;
							}
						} else
						{
							for(i=c->open->vbi_fmt.start[1]; i<(c->open->vbi_fmt.start[1]+c->open->vbi_fmt.count[1]); i++)
								btv->vbi_even_tab[i*2] = btv->vbi_dummy_phys;
						}
					}
				}
			}
			while(c_done)
			{
				v4l2_masterclock_gettime(&c_done->buf.timestamp);
				c_done->buf.flags |= V4L2_BUF_FLAG_DONE | V4L2_BUF_FLAG_KEYFRAME;
				c_done->buf.bytesused = odd; // just pass this here for now
				v4l2_q_add_tail(&c_done->c->cap_done_q, &c_done->q);
				wake_up_interruptible(&c_done->c->open->waitq);
				c_done = c_done->next;
			}
			if(btv->sync > 0) /* sync requested */
			{
				btv->sync = 0;
				Dprintk(2,"bttv%d: sending sync!\n", btv->nr);
				wake_up_interruptible(&btv->riscq);
			}
		}
		spin_unlock(&btv->s_lock);

		if (astat&BT848_INT_HLOCK) 
		{
		/*	if ((dstat&BT848_DSTATUS_HLOC) || (btv->radio))
				audio(btv, AUDIO_ON,0);
			else
				audio(btv, AUDIO_OFF,0); */
		}
    
		count++;
		if (count > 20) 
		{
			btwrite(0, BT848_INT_MASK);
			printk(KERN_ERR 
			       "bttv%d: IRQ lockup, cleared int mask\n", btv->nr);
			// bt848_offline(btv);
		}
	}
}

/*
 *	Scan for a Bt848 card, request the irq and map the io memory 
 */

static void __devexit bttv_remove(struct pci_dev *pci_dev)
{
        u8 command;
        struct bttv *btv = pci_get_drvdata(pci_dev);
        
	Vprintk(1, "bttv%d: unloading\n", btv->nr);

        /* unregister i2c_bus */
	if (0 == btv->i2c_rc)
		i2c_bit_del_bus(&btv->i2c_adap);

        /* turn off all capturing, DMA and IRQs */
        btand(~15, BT848_GPIO_DMA_CTL);

        /* first disable interrupts before unmapping the memory! */
        btwrite(0, BT848_INT_MASK);
        btwrite(~0x0UL,BT848_INT_STAT);
        btwrite(0x0, BT848_GPIO_OUT_EN);
        if (bttv_gpio)
        	bttv_gpio_tracking(btv,"cleanup");

        /* disable PCI bus-mastering */
        pci_read_config_byte(btv->dev, PCI_COMMAND, &command);
        /* Should this be &=~ ?? */
        command&=~PCI_COMMAND_MASTER;
        pci_write_config_byte(btv->dev, PCI_COMMAND, command);

        /* unmap and free memory */

// ********* Free any memory

	bt848_free_dma(btv);

        free_irq(btv->irq,btv);
        Dprintk(1,KERN_DEBUG "bt848_mem: 0x%p.\n", btv->bt848_mem);
        if (btv->bt848_mem)
                iounmap(btv->bt848_mem);

	v4l2_unregister_device(&btv->video_dev);
	v4l2_unregister_device(&btv->vbi_dev);
	if(radio[btv->nr]>0)
		v4l2_unregister_device(&btv->radio_dev);
	
        release_mem_region(pci_resource_start(btv->dev,0),
                           pci_resource_len(btv->dev,0));
        /* wake up any waiting processes
           because shutdown flag is set, no new processes (in this queue)
           are expected
        */
        btv->shutdown=1;
        wake_up(&btv->gpioq);
        
        pci_set_drvdata(pci_dev, NULL);
        return;
}


static int __devinit bttv_probe(struct pci_dev *dev, const struct pci_device_id *pci_id)
{
	int result;
	unsigned char command,lat;
	struct bttv *btv;
#if defined(__powerpc__)
        unsigned int cmd;
#endif
	printk(KERN_INFO "bttv: Bt8xx card found (%d).\n", bttv_num);
	
        btv=&bttvs[bttv_num];
        memset(btv, 0, sizeof(struct bttv));
        btv->dev=dev;
        btv->nr = bttv_num;
        btv->bt848_mem=NULL;
	btv->s_lock = SPIN_LOCK_UNLOCKED;
        init_waitqueue_head(&btv->riscq);
	init_waitqueue_head(&btv->gpioq);
	btv->shutdown=0;

        btv->id=dev->device;
        btv->irq=dev->irq;
	btv->bt848_adr=pci_resource_start(dev,0);
	if (pci_enable_device(dev))
		return -EIO;
	if (!request_mem_region(pci_resource_start(dev,0),
				pci_resource_len(dev,0),
				"bttv")) {
		return -EBUSY;
	}
        if (btv->id >= 878)
                btv->i2c_command = 0x83;                   
        else
                btv->i2c_command=(I2C_TIMING | BT848_I2C_SCL | BT848_I2C_SDA);

        pci_read_config_byte(dev, PCI_CLASS_REVISION, &btv->revision);
	pci_read_config_byte(dev, PCI_LATENCY_TIMER, &lat);
	printk(KERN_INFO "bttv%d: Bt%d (rev %d) at %02x:%02x.%x, ",
		bttv_num,btv->id, btv->revision, dev->bus->number,
		PCI_SLOT(dev->devfn),PCI_FUNC(dev->devfn));
	printk("irq: %d, latency: %d, memory: 0x%lx\n",
		btv->irq, lat, btv->bt848_adr);
	
	bttv_idcard(btv);
		
#if defined(__powerpc__)
        /* on OpenFirmware machines (PowerMac at least), PCI memory cycle */
        /* response on cards with no firmware is not enabled by OF */
        pci_read_config_dword(dev, PCI_COMMAND, &cmd);
        cmd = (cmd | PCI_COMMAND_MEMORY ); 
        pci_write_config_dword(dev, PCI_COMMAND, cmd);
#endif

#ifdef __sparc__
	btv->bt848_mem=(unsigned char *)btv->bt848_adr;
#else
	btv->bt848_mem=ioremap(btv->bt848_adr, 0x1000);
#endif
        
        /* clear interrupt mask */
	btwrite(0, BT848_INT_MASK);

        result = request_irq(btv->irq, bttv_irq,
                             SA_SHIRQ | SA_INTERRUPT,"bttv",(void *)btv);
        if (result==-EINVAL) 
        {
                printk(KERN_ERR "bttv%d: Bad irq number or handler\n",
                       bttv_num);
		goto fail;
        }
        if (result==-EBUSY)
        {
                printk(KERN_ERR "bttv%d: IRQ %d busy, change your PnP config in BIOS\n",bttv_num,btv->irq);
		goto fail;
        }
        if (result < 0) 
		goto fail;
        
        pci_set_master(dev);

        btv->triton1=triton1 ? BT848_INT_ETBF : 0;
        if (triton1 && btv->id >= 878) 
        {
                btv->triton1 = 0;
                printk("bttv: Enabling 430FX compatibilty for bt878\n");
                pci_read_config_byte(dev, BT878_DEVCTRL, &command);
                command|=BT878_EN_TBFX;
                pci_write_config_byte(dev, BT878_DEVCTRL, command);
                pci_read_config_byte(dev, BT878_DEVCTRL, &command);
                if (!(command&BT878_EN_TBFX)) 
                {
                        printk("bttv: 430FX compatibility could not be enabled\n");
			free_irq(btv->irq,btv);
			result = -1;
			goto fail;
                }
        }

	pci_set_drvdata(dev,btv);

	if(init_bt848(btv) < 0) {
		bttv_remove(dev);
		return -EIO;
	}
	bttv_num++;
	
        return 0;
	
 fail:
	release_mem_region(pci_resource_start(btv->dev,0),
			   pci_resource_len(btv->dev,0));
	return result;
}

static struct pci_device_id bttv_pci_tbl[] __devinitdata = {
        {PCI_VENDOR_ID_BROOKTREE, PCI_DEVICE_ID_BT848,
         PCI_ANY_ID, PCI_ANY_ID, 0, 0, 0},
	{PCI_VENDOR_ID_BROOKTREE, PCI_DEVICE_ID_BT849,
         PCI_ANY_ID, PCI_ANY_ID, 0, 0, 0},
	{PCI_VENDOR_ID_BROOKTREE, PCI_DEVICE_ID_BT878,
         PCI_ANY_ID, PCI_ANY_ID, 0, 0, 0},
	{PCI_VENDOR_ID_BROOKTREE, PCI_DEVICE_ID_BT879,
         PCI_ANY_ID, PCI_ANY_ID, 0, 0, 0},
        {0,}
};

MODULE_DEVICE_TABLE(pci, bttv_pci_tbl);

static struct pci_driver bttv_pci_driver = {
        name:     "bttv",
        id_table: bttv_pci_tbl,
        probe:    bttv_probe,
        remove:   bttv_remove,
};

int bttv_init_module(void)
{
	bttv_num = 0;

	printk(KERN_INFO "bttv: driver version %d.%d.%d loaded\n",
	       (BTTV_VERSION_CODE >> 16) & 0xff,
	       (BTTV_VERSION_CODE >> 8) & 0xff,
	       BTTV_VERSION_CODE & 0xff);

	handle_chipset();

	return pci_module_init(&bttv_pci_driver);
}

void bttv_cleanup_module(void)
{
	pci_unregister_driver(&bttv_pci_driver);
	return;
}

module_init(bttv_init_module);
module_exit(bttv_cleanup_module);

/*
 * Local variables:
 * c-basic-offset: 8
 * End:
 */
